# -*- coding: utf-8 -*-
# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *
# * [1] SciLifeLab, Stockholm University
# *
# * This program is free software: you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation, either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program.  If not, see <https://www.gnu.org/licenses/>.
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

import pyworkflow.tests as pwtests
import pyworkflow.mapper as pwmapper
import pyworkflow.protocol as pwprot


# TODO: this test seems not to be finished.
from pyworkflowtests.protocols import SleepingProtocol
from pyworkflowtests import Domain


class TestProtocolExecution(pwtests.BaseTest):
    
    @classmethod
    def setUpClass(cls):
        pwtests.setupTestOutput(cls)
    
    def test_StepExecutor(self):
        """Test the list with several Complex"""
        fn = self.getOutputPath("protocol.sqlite")
        print("Writing to db: %s" % fn)

        # Discover objects and protocols
        mapperDict = Domain.getMapperDict()

        # Check that the protocol has associated package
        mapper = pwmapper.SqliteMapper(fn, mapperDict)
        prot = SleepingProtocol(mapper=mapper, n=2,
                                workingDir=self.getOutputPath(''))
        domain = prot.getClassDomain()
        domain.printInfo()

        prot.setStepsExecutor(pwprot.StepExecutor(hostConfig=None))
        prot.run()
        mapper.commit()
        mapper.close()

        self.assertEqual(prot._steps[0].getStatus(), pwprot.STATUS_FINISHED)
        
        mapper2 = pwmapper.SqliteMapper(fn, mapperDict)
        prot2 = mapper2.selectById(prot.getObjId())
        
        self.assertEqual(prot.endTime.get(), prot2.endTime.get())



#!/usr/bin/env python

import sys
import os

from pyworkflow.project import Manager
import pyworkflow.utils as pwutils


def usage(error):
    print("""
    ERROR: %s
    
    Usage: fixlinks.py PROJECT SEARCH_DIR
        PROJECT: provide the project name to fix broken links in the imports.
        SEARCH_DIR: provide a directory where to look for the files.
        and fix the links.    
    """ % error)
    sys.exit(1)    


if len(sys.argv) != 3:
    usage("Incorrect number of input parameters")

projName = sys.argv[1]
searchDir = os.path.abspath(sys.argv[2])

# Create a new project
manager = Manager()

if not manager.hasProject(projName):
    usage("Nonexistent project: %s" % pwutils.red(projName))
    
if not os.path.exists(searchDir):
    usage("Nonexistent SEARCH_DIR: %s" % pwutils.red(searchDir))
    
project = manager.loadProject(projName)

project.fixLinks(searchDir)



#!/usr/bin/env python

import logging
import unittest

from pyworkflow import Config
from pyworkflow.utils import (getLineInFile, isInFile, ScipionLogger,
                              getLogConfiguration)
from pyworkflow.tests import BaseTest, setupTestOutput


# FIXME:Nacho
# Ok, Nacho and Airen, explain what you have to fix! :)

class TestLogs(BaseTest):
    
    @classmethod
    def setUpClass(cls):
        setupTestOutput(cls)        

    def testSimpleFileLog(self):
        import random
        logTestCode = random.randint(1, 100000)

        genLogFn = Config.SCIPION_LOG
        getLogConfiguration()
        log1 = logging.getLogger('pyworkflow.test.log.test_scipion_log')
        genInfoTest = 'Testing general info [%d]' % logTestCode
        genDebugTest = 'Testing general debug [%d]' % logTestCode
        genWarningTest = 'Testing general warning [%d]' % logTestCode
        genErrorTest = 'Testing general error [%d]' % logTestCode
        log1.info(genInfoTest)
        # log.debug(genDebugTest)
        log1.warning(genWarningTest)

        logFn = self.getOutputPath('fileLog.log')
        log2 = ScipionLogger(logFn)
        fileInfoTest = 'Not really info, just testing logger  [%d]' % logTestCode
        fileDebugTest = 'Not really debug, just testing logger  [%d]' % logTestCode
        fileWarningTest = 'Not really a warning, just testing logger [%d]' % logTestCode
        fileErrorTest = 'Not really an error, just testing logger [%d]' % logTestCode
        log2.info(fileInfoTest)
        # log.debug(fileDebugTest)
        log2.warning(fileWarningTest)
        log3 = logging.getLogger('pyworkflow.tests.log')
        log3.error(genErrorTest)
        
        log4 = ScipionLogger(logFn)
        log4.error(fileErrorTest)
        
        # Check general logs
        lineGenInfoTest = getLineInFile(genInfoTest, genLogFn)
        lineGenWarningTest = getLineInFile(genWarningTest, genLogFn)
        lineGenErrorTest = getLineInFile(genErrorTest, genLogFn)
        
        isFileInfoTest = isInFile(fileInfoTest, genLogFn)
        isFileWarningTest = isInFile(fileWarningTest, genLogFn)
        isFileErrorTest = isInFile(fileErrorTest, genLogFn)
        
        genLoggerChecked = True
        if lineGenInfoTest is None:
            print('General info log failed!!!')
            genLoggerChecked = False
        if lineGenWarningTest is None:
            print('General warning log failed!!!')
            genLoggerChecked = False
        if lineGenErrorTest is None:
            print('General error log failed!!!')
            genLoggerChecked = False
        
        if not((lineGenInfoTest < lineGenWarningTest) & (lineGenWarningTest < lineGenErrorTest)):
            print('General logs have an incorrect order!!!')
            genLoggerChecked = False
        
        if isFileInfoTest | isFileWarningTest | isFileErrorTest:
            print('File logs in general log!!!')
            genLoggerChecked = False
        
        # Check file logs
        lineFileInfoTest = getLineInFile(fileInfoTest, logFn)
        lineFileWarningTest = getLineInFile(fileWarningTest, logFn)
        lineFileErrorTest = getLineInFile(fileErrorTest, logFn)
        
        isGenInfoTest = isInFile(genInfoTest, logFn)
        isGenWarningTest = isInFile(genWarningTest, logFn)
        isGenErrorTest = isInFile(genErrorTest, logFn)
        
        fileLoggerChecked = True
        if lineFileInfoTest is None:
            print('File info log failed!!!')
            fileLoggerChecked = False
        if lineFileWarningTest is None:
            print('File warning log failed!!!')
            fileLoggerChecked = False
        if lineFileErrorTest is None:
            print('File error log failed!!!')
            fileLoggerChecked = False
        
        if not((lineFileInfoTest < lineFileWarningTest) & (lineFileWarningTest < lineFileErrorTest)):
            print('File logs have an incorrect order!!!')
            fileLoggerChecked = False
        
        if isGenInfoTest | isGenWarningTest | isGenErrorTest:
            print('General logs in file log!!!')
            fileLoggerChecked = False 

        self.assertTrue(genLoggerChecked & fileLoggerChecked)
        
        
if __name__ == '__main__':
    unittest.main()



# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
"""
This modules contains classes to store information about
execution hosts.
"""


import os
import sys
import json
from configparser import RawConfigParser
from collections import OrderedDict

import pyworkflow as pw
from pyworkflow.object import OrderedObject, String, Integer, Boolean

        
class HostConfig(OrderedObject):
    """ Main store the configuration for execution hosts. """
    
    def __init__(self, **kwargs):
        OrderedObject.__init__(self, **kwargs)
        self.label = String(kwargs.get('label', None))
        self.hostName = String(kwargs.get('hostName', None))
        self.userName = String()
        self.password = String()
        self.hostPath = String()
        self.mpiCommand = String()
        self.scipionHome = String()
        self.scipionConfig = String()
        self.address = String()
        self.queueSystem = QueueSystemConfig()
    
    def getLabel(self):
        return self.label.get()
    
    def getHostName(self):
        return self.hostName.get()
    
    def getUserName(self):
        return self.userName.get()
    
    def getPassword(self):
        return self.password.get()
    
    def getHostPath(self):
        return self.hostPath.get()
    
    def getSubmitCommand(self):
        return self.queueSystem.submitCommand.get()
    
    def getSubmitPrefix(self):
        return self.queueSystem.submitPrefix.get()

    def getCheckCommand(self):
        return self.queueSystem.checkCommand.get()

    def getCancelCommand(self):
        return self.queueSystem.cancelCommand.get()
    
    def isQueueMandatory(self):
        return self.queueSystem.mandatory.get()
    
    def getSubmitTemplate(self):
        return self.queueSystem.getSubmitTemplate()
    
    def getQueuesDefault(self):
        return self.queueSystem.queuesDefault
    
    def getMpiCommand(self):
        return self.mpiCommand.get()
    
    def getQueueSystem(self):
        return self.queueSystem

    def getJobDoneRegex(self):
        return self.queueSystem.jobDoneRegex.get()
    
    def setLabel(self, label):
        self.label.set(label)
    
    def setHostName(self, hostName):
        self.hostName.set(hostName)
    
    def setUserName(self, userName):
        self.userName.set(userName)
    
    def setPassword(self, password):
        self.password.set(password)
    
    def setHostPath(self, hostPath):
        self.hostPath.set(hostPath)
    
    def setMpiCommand(self, mpiCommand):
        self.mpiCommand.set(mpiCommand)  
        
    def setQueueSystem(self, queueSystem):
        self.queueSystem = queueSystem
        
    def getScipionHome(self):
        """ Return the path where Scipion is installed in 
        the host. This is useful when launching remote jobs.
        """ 
        return self.scipionHome.get()
    
    def setScipionHome(self, newScipionHome):
        self.scipionHome.set(newScipionHome)
        
    def getScipionConfig(self):
        """ From which file to read the configuration file in 
        this hosts. Useful for remote jobs.
        """
        return self.scipionConfig.get()
    
    def setScipionConfig(self, newConfig):
        self.scipionConfig.set(newConfig)
        
    def getAddress(self):
        return self.address.get()
    
    def setAddress(self, newAddress):
        return self.address.set(newAddress)

    @classmethod
    def writeBasic(cls, configFn):
        """ Write a very basic Host configuration for testing purposes. """
        with open(configFn, 'w') as f:
            f.write('[localhost]\nPARALLEL_COMMAND = '
                    'mpirun -np %%(JOB_NODES)d --map-by node %%(COMMAND)s\n')

    @classmethod
    def load(cls, hostsConf):
        """ Load several hosts from a configuration file.
        Return an dictionary with hostName -> hostConfig pairs.
        """
        # Read from users' config file. Raw to avoid interpolation of %: we expect %_
        cp = RawConfigParser(comment_prefixes=";")
        cp.optionxform = str  # keep case (stackoverflow.com/questions/1611799)
        hosts = OrderedDict()

        try:
            assert cp.read(hostsConf) != [], 'Missing file %s' % hostsConf

            for hostName in cp.sections():
                host = HostConfig(label=hostName, hostName=hostName)
                host.setHostPath(pw.Config.SCIPION_USER_DATA)

                # Helper functions (to write less)
                def get(var, default=None):
                    if cp.has_option(hostName, var):

                        value = cp.get(hostName, var)
                        # Rescue python2.7 behaviour: ## at the beginning of a line, means a single #.
                        # https://github.com/scipion-em/scipion-pyworkflow/issues/70
                        value = value.replace("\n##", "\n#")

                        # Keep compatibility: %_ --> %%
                        value = value.replace('%_(', '%(')

                        return value
                    else:
                        return default

                def getDict(var):
                    od = OrderedDict()

                    if cp.has_option(hostName, var):
                        for key, value in json.loads(get(var)).items():
                            od[key] = value

                    return od

                host.setScipionHome(get(pw.SCIPION_HOME_VAR, pw.Config.SCIPION_HOME))
                host.setScipionConfig(pw.Config.SCIPION_CONFIG)
                # Read the address of the remote hosts,
                # using 'localhost' as default for backward compatibility
                host.setAddress(get('ADDRESS', 'localhost'))
                host.mpiCommand.set(get('PARALLEL_COMMAND'))
                host.queueSystem = QueueSystemConfig()
                hostQueue = host.queueSystem  # shortcut
                hostQueue.name.set(get('NAME'))

                # If the NAME is not provided or empty
                # do no try to parse the rest of Queue parameters
                if hostQueue.hasName():
                    hostQueue.setMandatory(get('MANDATORY', 0))
                    hostQueue.submitPrefix.set(get('SUBMIT_PREFIX', ''))
                    hostQueue.submitCommand.set(get('SUBMIT_COMMAND'))
                    hostQueue.submitTemplate.set(get('SUBMIT_TEMPLATE'))
                    hostQueue.cancelCommand.set(get('CANCEL_COMMAND'))
                    hostQueue.checkCommand.set(get('CHECK_COMMAND'))
                    hostQueue.jobDoneRegex.set(get('JOB_DONE_REGEX'))
                    hostQueue.queues = getDict('QUEUES')
                    hostQueue.queuesDefault = getDict('QUEUES_DEFAULT')

                hosts[hostName] = host

            return hosts
        except Exception as e:
            sys.exit('Failed to read settings. The reported error was:\n  %s\n'
                     'To solve it, delete %s and run again.'
                     % (e, os.path.abspath(hostsConf)))


class QueueSystemConfig(OrderedObject):
    def __init__(self, **kwargs):
        OrderedObject.__init__(self, **kwargs)
        self.name = String()
        # Number of cores from which the queue is mandatory
        # 0 means no mandatory at all
        # 1 will force to launch all jobs through the queue
        self.mandatory = Integer()
        self.queues = None  # List for queue configurations
        self.submitCommand = String()
        # Allow to change the prefix of submission scripts
        # we used by default the ID.job, but in some clusters
        # the job script should start by a letter
        self.submitPrefix = String()
        self.checkCommand = String()
        self.cancelCommand = String()
        self.submitTemplate = String()
        self.jobDoneRegex = String()

    def hasName(self):
        return self.name.hasValue()
    
    def hasValue(self):
        return self.hasName() and len(self.queues)
    
    def getName(self):
        return self.name.get()
    
    def getMandatory(self):
        return self.mandatory.get()
    
    def getSubmitTemplate(self):
        return self.submitTemplate.get()
    
    def getSubmitCommand(self):
        return self.submitCommand.get()
    
    def getCheckCommand(self):
        return self.checkCommand.get()
    
    def getCancelCommand(self):
        return self.cancelCommand.get()
    
    def getQueues(self):
        return self.queues
    
    def setName(self, name):
        self.name.set(name)
    
    def setMandatory(self, mandatory):
        # This condition is to be backward compatible
        # when mandatory was a boolean
        # now it should use the number of CPU
        # that should force to use the queue
        if mandatory in ['False', 'false']:
            mandatory = 0
        elif mandatory in ['True', 'true']:
            mandatory = 1
            
        self.mandatory.set(mandatory)
    
    def setSubmitTemplate(self, submitTemplate):
        self.submitTemplate.set(submitTemplate)
    
    def setSubmitCommand(self, submitCommand):
        self.submitCommand.set(submitCommand)
    
    def setCheckCommand(self, checkCommand):
        self.checkCommand.set(checkCommand)
    
    def setCancelCommand(self, cancelCommand):
        self.cancelCommand.set(cancelCommand)

    def setJobDoneRegex(self, jobDoneRegex):
        self.jobDoneRegex.set(jobDoneRegex)
    
    def setQueues(self, queues):
        self.queues = queues
        
    def getQueueConfig(self, objId):
        if objId is not None and self.queues is not None:
            for queueConfig in self.queues:
                if objId == queueConfig.getObjId():
                    return queueConfig
        return None
        
        
# TODO: maybe deprecated
class QueueConfig(OrderedObject):
    def __init__(self, **kwargs):
        OrderedObject.__init__(self, **kwargs)
        self.name = String('default')
        self.maxCores = Integer()
        self.allowMPI = Boolean()
        self.allowThreads = Boolean()
        self.maxHours = Integer()
        
    def getName(self):
        return self.name.get()
    
    def getMaxCores(self):
        return self.maxCores.get()
    
    def getAllowMPI(self):
        return self.allowMPI.get()
    
    def getAllowThreads(self):
        return self.allowThreads.get()
    
    def getMaxHours(self):
        return self.maxHours.get()
    
    def setName(self, name):
        self.name.set(name)
    
    def setMaxCores(self, maxCores):
        self.maxCores.set(maxCores)
        
    def setAllowMPI(self, allowMPI):
        self.allowMPI.set(allowMPI)
    
    def setAllowThreads(self, allowThreads):
        self.allowThreads.set(allowThreads)
    
    def setMaxHours(self, maxHours):
        self.maxHours.set(maxHours)



# **************************************************************************
# *
# * Authors:     Yaiza Rancel (cyrancel@cnb.csic.es) [1]
# *              Pablo Conesa (pconesa@cnb.csic.es) [1]
# *              J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [2]
# *
# * [1] Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# * [2] SciLifeLab, Stockholm University
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
import glob
import os
import importlib
import inspect
import traceback
import types
import pkg_resources
from email import message_from_string
from collections import OrderedDict
from abc import ABCMeta, abstractmethod

import pyworkflow as pw
import pyworkflow.utils as pwutils
import pyworkflow.object as pwobj
from pyworkflow.template import Template

from .constants import *


class Domain:
    """ Class to represent the application domain.
    It will allow to specify new objects, protocols, viewers and wizards
    through the registration of new plugins.
    """

    # The following classes should be defined in subclasses of Domain
    _name = None
    _protocolClass = None
    _objectClass = pwobj.Object
    _viewerClass = None
    _wizardClass = None
    _baseClasses = {}  # Update this with the base classes of the Domain

    # Dictionaries to store different type of objects
    _plugins = {}
    _protocols = {}
    _objects = {}
    _viewers = {}
    _wizards = {}

    @classmethod
    def registerPlugin(cls, name):
        """ Register a new plugin. This function should only be called when
        creating a class with __metaclass__=PluginMeta that will trigger this.
        """
        m = importlib.import_module(name)

        # Define variables
        m.Plugin._defineVariables()
        m.Domain = cls  # Register the domain class for this module
        # TODO avoid loading bibtex here and make a lazy load like the rest.
        # Load bibtex
        m._bibtex = {}
        bib = cls.__getSubmodule(name, 'bibtex')
        if bib is not None:
            if hasattr(bib, "_bibtex"):
                print("WARNING FOR DEVELOPERS:  %s/%s._bibtex unnecessarily declared. Just the doc string is enough." % (name, "bibtex"))
            else:
                try:
                    m._bibtex = pwutils.LazyDict(lambda: pwutils.parseBibTex(bib.__doc__))
                except Exception:
                    pass
        cls._plugins[name] = m  # Register the name to as a plugin

    @classmethod
    def getPlugins(cls):
        """ Return existing plugins for this Domain. """
        loaded = getattr(cls, '_pluginsLoaded', False)
        if not loaded:  # Load plugin only once
            cls._discoverPlugins()
            cls._pluginsLoaded = True
        return dict(cls._plugins)

    @classmethod
    def _discoverPlugins(cls):
        for entry_point in pkg_resources.iter_entry_points('pyworkflow.plugin'):
            cls.registerPlugin(entry_point.name)

    @classmethod
    def _discoverGUIPlugins(cls):
        for entry_point in pkg_resources.iter_entry_points('pyworkflow.guiplugin'):
            entry_point.load()

    @classmethod
    def getPlugin(cls, name):
        """ Load a given plugin name. """
        m = importlib.import_module(name)

        # if not cls.__isPlugin(m):
        #     raise Exception("Invalid plugin '%s'. "
        #                     "Class Plugin with __metaclass__=PluginMeta "
        #                     "not found" % name)
        return m

    @classmethod
    def refreshPlugin(cls, name):
        """ Refresh a given plugin name. """
        plugin = cls.getPlugin(name)
        if plugin is not None:
            fn = plugin.__file__
            fn_dir = os.path.dirname(fn) + os.sep
            module_visit = {plugin}
            module_visit_path = {fn}
            del fn

            def get_recursive_modules(module):
                """ Get all plugin modules recursively """
                for module_child in vars(module).values():
                    if isinstance(module_child, types.ModuleType):
                        fn_child = getattr(module_child, "__file__", None)
                        if (fn_child is not None) and fn_child.startswith(
                                fn_dir):
                            if fn_child not in module_visit_path:
                                module_visit.add(module_child)
                                module_visit_path.add(fn_child)
                                get_recursive_modules(module_child)

            get_recursive_modules(plugin)
            # reload all plugin modules

            while module_visit:
                for module in module_visit:
                    try:
                        importlib.reload(module)
                        module_visit.remove(module)
                        break
                    except Exception as ex:
                        pass

    @classmethod
    def __getSubclasses(cls, submoduleName, BaseClass,
                        updateBaseClasses=False):
        """ Load all detected subclasses of a given BaseClass.
        Params:
            updateBaseClasses: if True, it will try to load classes from the
                Domain submodule that was not imported as globals()
        """
        subclasses = getattr(cls, '_%s' % submoduleName)

        if not subclasses:  # Only discover subclasses once
            if updateBaseClasses:
                sub = cls.__getSubmodule(cls.getName(), submoduleName)
                if sub is not None:
                    for name in cls.getModuleClasses(sub):
                        attr = getattr(sub, name)
                        if inspect.isclass(attr) and issubclass(attr, BaseClass):
                            cls._baseClasses[name] = attr

            for pluginName, plugin in cls.getPlugins().items():
                sub = cls.__getSubmodule(pluginName, submoduleName)
                if sub is not None:
                    for name in cls.getModuleClasses(sub):
                        attr = getattr(sub, name)
                        if inspect.isclass(attr) and issubclass(attr, BaseClass):

                            # Check if the class already exists (to prevent
                            # naming collisions)
                            if name in subclasses:
                                # Get already added class plugin
                                pluginCollision = subclasses[name]._package.__name__
                                print("ERROR: Name collision (%s) detected "
                                      "while discovering %s.%s.\n"
                                      " It conflicts with %s" %
                                      (name, pluginName, submoduleName,
                                       pluginCollision))
                            else:
                                # Set this special property used by Scipion
                                attr._package = plugin
                                subclasses[name] = attr
            subclasses.update(
                pwutils.getSubclasses(BaseClass, cls._baseClasses))

        return subclasses

    @classmethod
    def getModuleClasses(cls, module):
        # Dir was used before but dir returns all imported elements
        # included those imported to be BaseClasses.
        # return dir(module)

        # Get the module name
        moduleName = module.__name__

        # Get any module class
        for name, declaredClass in inspect.getmembers(module, inspect.isclass):
            if moduleName in declaredClass.__module__:
                yield name

    @classmethod
    def getProtocols(cls):
        """ Return all Protocol subclasses from all plugins for this domain."""
        return cls.__getSubclasses('protocols', cls._protocolClass)

    @classmethod
    def getObjects(cls):
        """ Return all EMObject subclasses from all plugins for this domain."""
        return cls.__getSubclasses('objects', cls._objectClass)

    @classmethod
    def getViewers(cls):
        """ Return all Viewer subclasses from all plugins for this domain."""
        return cls.__getSubclasses('viewers', cls._viewerClass,
                                   updateBaseClasses=True)

    @classmethod
    def getWizards(cls):
        """ Return all Wizard subclasses from all plugins for this domain."""
        return cls.__getSubclasses('wizards', cls._wizardClass)

    @classmethod
    def getMapperDict(cls):
        """ Return a dictionary that can be used with subclasses of Mapper
        to store/retrieve objects (including protocols) defined in this
        Domain. """
        mapperDict = getattr(cls, '__mapperDict', None)

        if mapperDict is None:
            mapperDict = dict(pwobj.OBJECTS_DICT)
            mapperDict.update(cls.getObjects())
            mapperDict.update(cls.getProtocols())
            cls.__mapperDict = mapperDict

        return mapperDict

    @classmethod
    def getName(cls):
        """ Return the name of this Domain. """
        return cls._name

    @staticmethod
    def importFromPlugin(module, objects=None, errorMsg='', doRaise=False):
        """ This method try to import either a list of objects from the
            module/plugin or the whole module/plugin and returns what is
            imported if not fails.
            When the import fails (due to the plugin or the object is not found),
            it prints a common message + optional errorMsg;
            or it raise an error with the same message, if doRaise is True.

         -> Usages:

             # Import the whole plugin 'plugin1' as 'plug1'
             plug1 = importFromPlugin('plugin1')

             # Import a plugin's module
             pl1Cons = importFromPlugin('plug1.constants')

             # Import a single class from a plugin's module
             p1prot1 = importFromPlugin('plug1.protocols', 'prot1')

             # Import some classes from a plugin's module,
             #   the returned tuple has the same length of the second argument
             pt1, pt2, ... = importFromPlugin('plugin1.protocols',
                                              ['pt1', 'pt2', ...])
        """
        def _tryImportFromPlugin(submodule=None):
            try:
                if submodule is None:  # Import only the module
                    output = importlib.import_module(module)
                else:  # Import the class of that module
                    output = getattr(importlib.import_module(module), submodule)
                return output
            except Exception as e:
                plugName = module.split('.')[0]  # The Main module is the plugin
                errMsg = (str(e) if errorMsg == ''
                          else "%s. %s" % (str(e), errorMsg))
                Domain.__pluginNotFound(plugName, errMsg, doRaise)

        if objects is None or isinstance(objects, str):
            output = _tryImportFromPlugin(objects)
        else:
            output = tuple()
            for obj in objects:
                output += (_tryImportFromPlugin(obj), )  # append in tuple
        return output

    @classmethod
    def findClass(cls, className):
        """ Find a class object given its name.
        The search will start with protocols and then with protocols.
        """
        # FIXME: Why not also Viewers, Wizards?
        c = cls.getProtocols().get(className,
                                   cls.getObjects().get(className, None))
        if c is None:
            raise Exception("findClass: class '%s' not found." % className)
        return c

    @classmethod
    def findSubClasses(cls, classDict, className):
        """ Find all subclasses of a give className. """
        clsObj = classDict[className]
        subclasses = {}

        for k, v in classDict.items():
            if issubclass(v, clsObj):
                subclasses[k] = v
        return subclasses

    @classmethod
    def getPreferredViewers(cls, className):
        """ Find and import the preferred viewers for this class. """
        viewerNames = pw.Config.VIEWERS.get(className, [])
        if not isinstance(viewerNames, list):
            viewerNames = [viewerNames]
        viewers = []  # we will try to import them and store here
        for prefViewerStr in viewerNames:
            try:
                viewerModule, viewerClassName = prefViewerStr.rsplit('.', 1)
                prefViewer = cls.importFromPlugin(viewerModule,
                                                  viewerClassName,
                                                  doRaise=True)
                viewers.append(prefViewer)
            except Exception as e:
                print("Couldn't load \"%s\" as preferred viewer.\n"
                      "There might be a typo in your VIEWERS variable "
                      "or an error in the viewer's plugin installation"
                      % prefViewerStr)
                print(e)
        return viewers

    @classmethod
    def findViewers(cls, className, environment):
        """ Find the available viewers in this Domain for this class. """
        viewers = []
        try:
            clazz = cls.findClass(className)
            baseClasses = clazz.mro()
            preferredViewers = cls.getPreferredViewers(className)
            preferedFlag = 0

            for viewer in cls.getViewers().values():
                if environment in viewer._environments:
                    for t in viewer._targets:
                        if t in baseClasses:
                            for prefViewer in preferredViewers:
                                if viewer is prefViewer:
                                    viewers.insert(0, viewer)
                                    preferedFlag = 1
                                    break
                            else:
                                if t == clazz:
                                    viewers.insert(preferedFlag, viewer)
                                else:
                                    viewers.append(viewer)
                                break
        except Exception as e:
            # Catch if there is a missing plugin, we will get Legacy which triggers and Exception
            pass

        return viewers

    @classmethod
    def findWizards(cls, protocol, environment):
        """ Find available wizards for this class, in this Domain.
        Params:
            protocols: Protocol instance for which wizards will be search.
            environment: The environment name for wizards (e.g TKINTER)
        Returns:
            a dict with the paramName and wizards for this class."""
        return cls.__findWizardsFromDict(protocol, environment,
                                         cls.getWizards())

    @classmethod
    def printInfo(cls):
        """ Simple function (mainly for debugging) that prints basic
        information about this Domain. """
        print("Domain: %s" % cls._name)
        print("     objects: %s" % len(cls._objects))
        print("   protocols: %s" % len(cls._protocols))
        print("     viewers: %s" % len(cls._viewers))
        print("     wizards: %s" % len(cls._wizards))

    # ---------- Private methods of Domain class ------------------------------
    @staticmethod
    def __pluginNotFound(plugName, errorMsg='', doRaise=False):
        """ Prints or raise (depending on the doRaise) telling why it is failing
        """
        hint = ("   Check the plugin manager (Configuration->Plugins in "
                "Scipion manager window) \n")
        # the str casting is to work with Exceptions as errorMsg
        if 'No module named %s' % plugName in str(errorMsg):
            msgStr = " > %s plugin not found. %s" % (plugName, errorMsg)
            hint += "   or use 'scipion installp --help' in the command line "
            hint += "to install it."
        else:
            msgStr = " > error when importing from %s: %s" % (plugName, errorMsg)
            if errorMsg != '':  # if empty we know nothing...
                hint += ("   or use 'scipion installp --help --checkUpdates' "
                         "in the command line to check for upgrades,\n   "
                         "it could be a versions compatibility issue.")

        stackList = traceback.extract_stack()
        if len(stackList) > 3:
            callIdx = -3  # We use the most probable index as default
            for idx, stackLine in enumerate(stackList):
                if stackLine[0].endswith('/unittest/loader.py'):
                    callIdx = idx + 1
        else:
            callIdx = 0

        callBy = stackList[callIdx][0]
        if callBy.endswith('pyworkflow/plugin.py'):
            # This special case is to know why is failing and not where is called
            # because we know that we call all plugins.protocols at the beginning
            calling = traceback.format_exc().split('\n')[-4]
        else:
            line = stackList[callIdx][1]
            calling = "  Called by %s, line %s" % (callBy, line)

        raiseMsg = "%s\n %s\n%s\n" % (msgStr, calling, hint)
        if doRaise:
            raise Exception("\n\n" + raiseMsg)
        else:
            print(raiseMsg)

    @staticmethod
    def __getSubmodule(name, subname):
        try:
            completeModuleText = '%s.%s' % (name, subname)
            if pwutils.isModuleAFolder(name):
                return importlib.import_module(completeModuleText)
        except Exception as e:
            msg = str(e)
            # FIXME: The following is a quick and dirty way to filter
            # when the submodule is not present
            if msg != 'No module named \'%s\'' % completeModuleText:
                Domain.__pluginNotFound(completeModuleText, msg)
            return None

    @classmethod
    def __isPlugin(cls, m):
        """ Return True if the module is a Scipion plugin. """
        return m.__name__ in cls._plugins

    @staticmethod
    def __findWizardsFromDict(protocol, environment, wizDict):
        wizards = {}
        baseClasses = [c.__name__ for c in protocol.getClass().mro()]

        for wiz in wizDict.values():
            if environment in wiz._environments:
                for c, params in wiz._targets:
                    if c.__name__ in baseClasses:
                        for p in params:
                            wizards[p] = wiz
        return wizards


class Plugin:
    __metaclass__ = ABCMeta

    _vars = {}
    _homeVar = ''  # Change in subclasses to define the "home" variable
    _pathVars = []
    _supportedVersions = []
    _name = ""
    _url = ""  # For the plugin
    _condaActivationCmd = None

    @classmethod
    def _defineVar(cls, varName, defaultValue):
        """ Internal method to define variables trying to get it from the environment first. """
        cls._addVar(varName, os.environ.get(varName, defaultValue))

    @classmethod
    def _addVar(cls, varName, value):
        """ Adds a variable to the local variable dictionary directly. Avoiding the environment"""
        cls._vars[varName] = value

    @classmethod
    @abstractmethod
    def getEnviron(cls):
        """ Setup the environment variables needed to launch programs. """
        pass

    @classmethod
    def getCondaActivationCmd(cls):
        if cls._condaActivationCmd is None:
            condaActivationCmd = os.environ.get(CONDA_ACTIVATION_CMD_VAR, "")
            if not condaActivationCmd:
                print("WARNING!!_condaActivationCmd: %s variable not defined. "
                      "Relying on conda being in the PATH" % CONDA_ACTIVATION_CMD_VAR)
            elif condaActivationCmd[-1] not in [";", "&"]:
                condaActivationCmd += "&&"

            cls._condaActivationCmd = condaActivationCmd

        return cls._condaActivationCmd

    @classmethod
    @abstractmethod
    def _defineVariables(cls):
        """ Method to define variables and their default values.
        It will use the method _defineVar that will take a variable value
        from the environment or from an optional default value.

        This method is not supposed to be called from client code,
        except from the Domain class when registering a Plugin.
        """
        pass

    @classmethod
    @abstractmethod
    def defineBinaries(cls, env):
        """ Define required binaries in the given Environment. """
        pass

    @classmethod
    def getVar(cls, varName, defaultValue=None):
        """ Return the value of a given variable. """
        return cls._vars.get(varName, defaultValue)

    @classmethod
    def getVars(cls):
        """ Return the value of a given variable. """
        return cls._vars

    @classmethod
    def getHome(cls, *paths):
        """ Return a path from the "home" of the package
         if the _homeVar is defined in the plugin. """
        home = cls.getVar(cls._homeVar)
        return os.path.join(home, *paths) if home else ''

    @classmethod
    def getSupportedVersions(cls):
        """ Return the list of supported binary versions. """
        return cls._supportedVersions

    @classmethod
    def getActiveVersion(cls, home=None, versions=None):
        """ Return the version of the binaries that are currently active.
        In the current implementation it will be inferred from the *_HOME
        variable, so it should contain the version number in it. """
        # FIXME: (JMRT) Use the basename might alleviate the issue with matching
        # the binaries version, but we might consider to find a better solution
        home = os.path.basename(home or cls.getHome())
        versions = versions or cls.getSupportedVersions()

        for v in versions:
            if v in home:
                return v

        return ''

    @classmethod
    def getName(cls):
        return cls.__module__

    @classmethod
    def validateInstallation(cls):
        """
        Check if the binaries are properly installed and if not, return
        a list with the error messages.

        The default implementation will check if the _pathVars exists.
        """
        try:
            missing = ["%s: %s" % (var, cls.getVar(var))
                       for var in cls._pathVars if not os.path.exists(cls.getVar(var))]

            return (["Missing variables:"] + missing) if missing else []
        except Exception as e:
            return ["validateInstallation fails: %s" % e]

    @classmethod
    def getPluginTemplateDir(cls):
        return os.path.join(pw.getModuleFolder(cls.getName()), 'templates')

    @classmethod
    def getTemplates(cls):
        """ Get the plugin templates from the templates directory.
            If more than one template is found or passed, a dialog is raised
            to choose one.
        """
        tempList = []
        pluginName = cls.getName()
        tDir = cls.getPluginTemplateDir()
        if os.path.exists(tDir):
            for file in glob.glob1(tDir, "*" + SCIPION_JSON_TEMPLATES):
                t = Template(pluginName, os.path.join(tDir, file))
                tempList.append(t)

        return tempList

    @classmethod
    def getUrl(cls, protClass=None):
        """ Url for the plugin to point users to it"""
        return cls._url


class PluginInfo:
    """
    Information related to a given plugin when it is installed via PIP
    """
    def __init__(self, name):
        try:
            dist = pkg_resources.get_distribution(name)
            lines = [l for l in dist._get_metadata(dist.PKG_INFO)]
            tuples = message_from_string('\n'.join(lines))

        except Exception:
            print("Plugin %s seems is not a pip module yet. "
                  "No metadata found" % name)
            tuples = message_from_string('Author: plugin in development mode?')

        self._name = name
        self._metadata = OrderedDict()

        for v in tuples.items():
            if v[0] == 'Keywords':
                break
            self._metadata[v[0]] = v[1]

    def getAuthor(self):
        return self._metadata.get('Author', "")

    def getAuthorEmail(self):
        return self._metadata.get('Author-email', '')

    def getHomePage(self):
        return self._metadata.get('Home-page', '')

    def getKeywords(self):
        return self._metadata.get('Keywords', '')



#!/usr/bin/env python
# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *
# * Unidad de Bioinformatica of Centro Nacional de Biotecnologia, CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import sys
import os

import pyworkflow as pw
from pyworkflow.project import Manager
from pyworkflow.object import Boolean


def usage(error):
    print("""
    ERROR: %s
    
    Usage: scipion python scripts/config_projects.py ProjectName [readOnly=True|False] [lifeTime=X|None]
        This script show (or edit) some of the configuration of the project.
        Use readOnly=True (or False) to set/unset read only property
        Use lifeTime=X for setting X hours or None to unset life time of the project.
    """ % error)
    sys.exit(1)    


n = len(sys.argv)

if n < 2 or n > 4:
    usage("Incorrect number of input parameters")
    
# Load the given project
projectsDir = os.path.join(pw.Config.SCIPION_USER_DATA, 'projects')
projName = sys.argv[1]
manager = Manager()
project = manager.loadProject(projName)

if project is None:
    usage("Project '%s' does not exist in: \n  %s" % (projName, projectsDir))

setReadOnly = False
setLifeTime = False

for arg in sys.argv:
    if arg.startswith('readOnly='):
        setReadOnly = True
        value = arg.split('readOnly=')[1]
        b = Boolean(value=value)
        readOnlyValue = b.get()
    elif arg.startswith('lifeTime='):
        setLifeTime = True
        value = arg.split('lifeTime=')[1]
        lifeTimeValue = None if value == 'None' else int(value)

if setReadOnly:
    project.setReadOnly(readOnlyValue)
    
if setLifeTime:
    project.settings.setLifeTime(lifeTimeValue)
    
if setReadOnly or setLifeTime:
    # Truly write settings
    project.settings.write()

print("Projects: ", projectsDir)
print("Project name: ", projName)
print(" Settings: ")
print("   readOnly = %s" % project.isReadOnly())
print("   lifeTime = %s (hours)" % project.settings.getLifeTime())




# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *
# * [1] SciLifeLab, Stockholm University
# *
# * This program is free software: you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation, either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program.  If not, see <https://www.gnu.org/licenses/>.
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
import os
import tkinter as tk
import tkinter.font as tkFont
import queue

from pyworkflow.object import OrderedObject
import pyworkflow as pw
from pyworkflow.utils import Message, Color, Icon

from .widgets import Button

# --------------- GUI CONFIGURATION parameters -----------------------
# TODO: read font size and name from config file
FONT_ITALIC = 'fontItalic'
FONT_NORMAL = 'fontNormal'
FONT_BOLD = 'fontBold'
FONT_BIG = 'fontBig'
cfgFontName = pw.Config.SCIPION_FONT_NAME
cfgFontSize = pw.Config.SCIPION_FONT_SIZE
cfgFontBigSize = cfgFontSize + 8
# TextColor
cfgCitationTextColor = "dark olive green"
cfgLabelTextColor = "black"
cfgSectionTextColor = "blue4"
# Background Color
cfgBgColor = "light grey"
cfgLabelBgColor = "white"
cfgHighlightBgColor = cfgBgColor
cfgButtonFgColor = "white"
cfgButtonActiveFgColor = "white"
cfgButtonBgColor = Color.RED_COLOR
cfgButtonActiveBgColor = "#A60C0C"
cfgEntryBgColor = "lemon chiffon"
cfgExpertLabelBgColor = "light salmon"
cfgSectionBgColor = cfgButtonBgColor
# Color
cfgListSelectColor = "DeepSkyBlue4"
cfgBooleanSelectColor = "white"
cfgButtonSelectColor = "DeepSkyBlue2"
# Dimensions limits
cfgMaxHeight = 650
cfgMaxWidth = 800
cfgMaxFontSize = 14
cfgMinFontSize = 6
cfgWrapLenght = cfgMaxWidth - 50


class Config(OrderedObject):
    pass


def saveConfig(filename):
    from pyworkflow.mapper import SqliteMapper
    from pyworkflow.object import String, Integer

    mapper = SqliteMapper(filename)
    o = Config()
    for k, v in globals().items():
        if k.startswith('cfg'):
            if type(v) is str:
                value = String(v)
            else:
                value = Integer(v)
            setattr(o, k, value)
    mapper.insert(o)
    mapper.commit()


# --------------- FONT related variables and functions  -----------------------
def setFont(fontKey, update=False, **opts):
    """Register a tkFont and store it in a globals of this module
    this method should be called only after a tk.Tk() windows has been
    created."""
    if not hasFont(fontKey) or update:
        globals()[fontKey] = tkFont.Font(**opts)

    return globals()[fontKey]


def hasFont(fontKey):
    return fontKey in globals()


def aliasFont(fontAlias, fontKey):
    """Set a fontAlias as another alias name of fontKey"""
    g = globals()
    g[fontAlias] = g[fontKey]


def getDefaultFont():
    return tk.font.nametofont("TkDefaultFont")


def getNamedFont(fontName):
    return globals()[fontName]


def getBigFont():
    return getNamedFont(FONT_BIG)


def setCommonFonts(windows=None):
    """Set some predefined common fonts.
    Same conditions of setFont applies here."""
    f = setFont(FONT_NORMAL, family=cfgFontName, size=cfgFontSize)
    aliasFont('fontButton', FONT_NORMAL)

    # Set default font size
    default_font = getDefaultFont()
    default_font.configure(size=cfgFontSize, family=cfgFontName)

    fb = setFont(FONT_BOLD, family=cfgFontName, size=cfgFontSize,
                 weight='bold')
    fi = setFont(FONT_ITALIC, family=cfgFontName, size=cfgFontSize,
                 slant='italic')

    setFont(FONT_BIG, family=cfgFontName, size=cfgFontBigSize)

    # not used?
    # setFont('fontLabel', family=cfgFontName, size=cfgFontSize+1, weight='bold')

    if windows:
        windows.fontBig = tkFont.Font(size=cfgFontSize + 2, family=cfgFontName,
                                      weight='bold')
        windows.font = f
        windows.fontBold = fb
        windows.fontItalic = fi

        # This adds the default value for the listbox inside a combo box
        # Which seems to not react to default font!!
        windows.root.option_add("*TCombobox*Listbox*Font", default_font)
        windows.root.option_add("*TCombobox*Font", default_font)


def changeFontSizeByDeltha(font, deltha, minSize=-999, maxSize=999):
    size = font['size']
    new_size = size + deltha
    if minSize <= new_size <= maxSize:
        font.configure(size=new_size)


def changeFontSize(font, event, minSize=-999, maxSize=999):
    deltha = 2
    if event.char == '-':
        deltha = -2
    changeFontSizeByDeltha(font, deltha, minSize, maxSize)


# --------------- IMAGE related variables and functions -----------------------
def getImage(imageName, imgDict=None, tkImage=True, percent=100,
             maxheight=None):
    """ Search for the image in the RESOURCES path list. """
    if imageName is None:
        return None
    if imgDict is not None and imageName in imgDict:
        return imgDict[imageName]
    if not os.path.isabs(imageName):
        imagePath = pw.findResource(imageName)
    else:
        imagePath = imageName
    image = None
    if imagePath:
        from PIL import Image
        image = Image.open(imagePath)
        w, h = image.size
        newSize = None
        if percent != 100:  # Display image with other dimensions
            fp = float(percent) / 100.0
            newSize = int(fp * w), int(fp * h)
        elif maxheight and h > maxheight:
            newSize = int(w * float(maxheight) / h), maxheight
        if newSize:
            image.thumbnail(newSize, Image.ANTIALIAS)
        if tkImage:
            from PIL import ImageTk
            image = ImageTk.PhotoImage(image)
        if imgDict is not None:
            imgDict[imageName] = image
    return image


# ---------------- Windows geometry utilities -----------------------
def getGeometry(win):
    """ Return the geometry information of the windows
    It will be a tuple (width, height, x, y)
    """
    return (win.winfo_reqwidth(), win.winfo_reqheight(),
            win.winfo_x(), win.winfo_y())


def centerWindows(root, dim=None, refWindows=None):
    """Center a windows in the middle of the screen 
    or in the middle of other windows(refWindows param)"""
    root.update_idletasks()
    if dim is None:
        gw, gh, gx, gy = getGeometry(root)
    else:
        gw, gh = dim
    if refWindows:
        rw, rh, rx, ry = getGeometry(refWindows)
        x = rx + (rw - gw) / 2
        y = ry + (rh - gh) / 2
    else:
        w = root.winfo_screenwidth()
        h = root.winfo_screenheight()
        x = (w - gw) / 2
        y = (h - gh) / 2

    root.geometry("%dx%d+%d+%d" % (gw, gh, x, y))


def configureWeigths(widget, row=0, column=0):
    """This function is a shortcut to a common
    used pair of calls: rowconfigure and columnconfigure
    for making childs widgets take the space available"""
    widget.columnconfigure(column, weight=1)
    widget.rowconfigure(row, weight=1)


class Window:
    """Class to manage a Tk windows.
    It will encapsulates some basic creation and 
    setup functions. """
    # To allow plugins to add their own menus
    _pluginMenus = list()

    def __init__(self, title='', masterWindow=None, weight=True,
                 minsize=(500, 300), icon=Icon.SCIPION_ICON, **kwargs):
        """Create a Tk window.
        title: string to use as title for the windows.
        master: if not provided, the windows create will be the principal one
        weight: if true, the first col and row will be configured with weight=1
        minsize: a minimum size for height and width
        icon: if not None, set the windows icon
        """
        # Init gui plugins
        pw.Config.getDomain()._discoverGUIPlugins()

        if masterWindow is None:
            Window._root = self
            self._images = {}
            # If a window which isn't the main Scipion window is generated from another main window, e. g. with Scipion
            # template after the refactoring of the kickoff, in which a dialog is launched and then a form, being it
            # called from the command line, so there's no Scipion main window. In that case, a tk.Tk() exists because if
            # a tk.TopLevel(), as the dialog, is directly launched, it automatically generates a main tk.Tk(). Thus,
            # after that first auto-tk.Tk(), another tk.Tk() was created here, and so the previous information was lost.
            # Solution proposed is to generate the root as an invisible window if it doesn't exist previously, and make
            # he first window generated a tk.Toplevel. After that, all steps executed later will go through the else
            # statement, being that way each new tk.Toplevel() correctly referenced.
            tk.Tk().withdraw()  # Main window, invisible
            self.root = tk.Toplevel()  # Toplevel of main window
        else:
            self.root = tk.Toplevel(masterWindow.root)
            self._images = masterWindow._images

        self.root.withdraw()
        self.root.title(title)

        if weight:
            configureWeigths(self.root)
        if minsize is not None:
            self.root.minsize(minsize[0], minsize[1])

        # Set the icon
        self._setIcon(icon)

        self.root.protocol("WM_DELETE_WINDOW", self._onClosing)
        self._w, self._h, self._x, self._y = 0, 0, 0, 0
        self.root.bind("<Configure>", self._configure)
        self.master = masterWindow
        setCommonFonts(self)

        if kwargs.get('enableQueue', False):
            self.queue = queue.Queue(maxsize=0)
        else:
            self.queue = None

    def _setIcon(self, icon):

        if icon is not None:
            try:
                path = pw.findResource(icon)
                # If path is None --> Icon not found
                if path is None:
                    # By default, if icon is not found use default scipion one.
                    path = pw.findResource(Icon.SCIPION_ICON)

                abspath = os.path.abspath(path)

                img = tk.Image("photo", file=abspath)
                self.root.tk.call('wm', 'iconphoto', self.root._w, img)
            except Exception as e:
                # Do nothing if icon could not be loaded
                pass

    def __processQueue(self):  # called from main frame
        if not self.queue.empty():
            func = self.queue.get(block=False)
            # executes graphic interface function
            func()
        self._queueTimer = self.root.after(500, self.__processQueue)

    def enqueue(self, func):
        """ Put some function to be executed in the GUI main thread. """
        self.queue.put(func)

    def getRoot(self):
        return self.root

    def desiredDimensions(self):
        """Override this method to calculate desired dimensions."""
        return None

    def _configure(self, e):
        """ Filter event and call appropriate handler. """
        if self.root != e.widget:
            return

        _, _, x, y = getGeometry(self.root)
        w, h = e.width, e.height

        if w != self._w or h != self._h:
            self._w, self._h = w, h
            self.handleResize()

        if x != self._x or y != self._y:
            self._x, self._y = x, y
            self.handleMove()

    def handleResize(self):
        """Override this method to respond to resize events."""
        pass

    def handleMove(self):
        """Override this method to respond to move events."""
        pass

    def show(self, center=True):
        """This function will enter in the Tk mainloop"""
        if center:
            if self.master is None:
                refw = None
            else:
                refw = self.master.root
            centerWindows(self.root, dim=self.desiredDimensions(),
                          refWindows=refw)
        self.root.deiconify()
        self.root.focus_set()
        if self.queue is not None:
            self._queueTimer = self.root.after(1000, self.__processQueue)
        self.root.mainloop()

    def close(self, e=None):
        self.root.destroy()
        # JMRT: For some reason when Tkinter has an exception
        # it does not exit the application as expected and
        # remains in the mainloop, so here we are forcing
        # to exit the whole system (only applies for the main window)
        if self.master is None:
            import sys
            sys.exit()

    def _onClosing(self):
        """Do some cleaning before closing."""
        if self.master is None:
            pass
        else:
            self.master.root.focus_set()
        if self.queue is not None:
            self.root.after_cancel(self._queueTimer)
        self.close()

    def getImage(self, imgName, percent=100, maxheight=None):
        return getImage(imgName, self._images, percent=percent,
                        maxheight=maxheight)

    def createMainMenu(self, menuConfig):
        """Create Main menu from the given MenuConfig object."""
        menu = tk.Menu(self.root, font=self.font)
        self._addMenuChilds(menu, menuConfig)
        self._addPluginMenus(menu)
        self.root.config(menu=menu)
        return menu

    def _addMenuChilds(self, menu, menuConfig):
        """Add entries of menuConfig in menu
        (using add_cascade or add_command for sub-menus and final options)."""
        # Helper function to create the main menu.
        for sub in menuConfig:
            menuLabel = sub.text.get()
            if not menuLabel:  # empty or None label means a separator
                menu.add_separator()
            elif len(sub) > 0:  # sub-menu
                submenu = tk.Menu(self.root, tearoff=0, font=self.font)
                menu.add_cascade(label=menuLabel, menu=submenu)
                self._addMenuChilds(submenu, sub)  # recursive filling
            else:  # menu option
                # If there is an entry called "Browse files", when clicked it
                # will call the method onBrowseFiles() (it has to be defined!)
                def callback(name):
                    """Return a callback function named "on<Name>"."""
                    f = "on%s" % "".join(x.capitalize() for x in name.split())
                    return lambda: getattr(self, f)()

                if sub.shortCut.get() is not None:
                    menuLabel += ' (' + sub.shortCut.get() + ')'

                menu.add_command(label=menuLabel, compound=tk.LEFT,
                                 image=self.getImage(sub.icon.get()),
                                 command=callback(name=sub.text.get()))

    def _addPluginMenus(self, menu):

        if self._pluginMenus:
            submenu = tk.Menu(self.root, tearoff=0, font=self.font)
            menu.add_cascade(label="Others", menu=submenu)

            # For each plugin menu
            for label, callback, icon in self._pluginMenus:
                submenu.add_command(label=label, compound=tk.LEFT,
                                    image=self.getImage(icon),
                                    command=callback)

    @classmethod
    def registerPluginMenu(cls, label, callback, icon=None):
        # TODO: have a proper model instead of a tuple?
        cls._pluginMenus.append((label, callback, icon))

    def showError(self, msg, header="Error", exception=None):
        """Pops up a dialog with the error message
        :param msg Message to display
        :param header Title of the dialog
        :param exception: Optional. exception associated"""
        from .dialog import showError
        showError(header, msg, self.root, exception=exception)

    def showInfo(self, msg, header="Info"):
        from .dialog import showInfo
        showInfo(header, msg, self.root)

    def showWarning(self, msg, header='Warning'):
        from .dialog import showWarning
        showWarning(header, msg, self.root)

    def askYesNo(self, title, msg):
        from .dialog import askYesNo
        return askYesNo(title, msg, self.root)

    def createCloseButton(self, parent):
        """ Create a button for closing the window, setting
        the proper label and icon. 
        """
        return Button(parent, Message.LABEL_BUTTON_CLOSE, Icon.ACTION_CLOSE,
                      command=self.close)

    def configureWeights(self, row=0, column=0):
        configureWeigths(self.root, row, column)



# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
from pyworkflow import Config


class GraphLayout(object):
    """ Base class for all algorithm that implement
    functions to organize a graph in a plane.
    """
    
    def draw(self, graph, **kwargs):
        """ Setup the nodes position in the plane. """
        pass


class BasicLayout(GraphLayout):
    """ This layout will keep node position as much as possible.
    It will try to allocate the nodes with x=0 and y=0.
    """
    
    def __init__(self):
        GraphLayout.__init__(self)
        self.DY = 65
        self.DX = 15
        
    def draw(self, graph, **kwargs):
        """ Organize nodes of the graph in the plane.
        Nodes should have: x, y, width and height attributes
        x and y will be modified.
        """
        for node in graph.getNodes():
            if hasattr(node, 'x') and hasattr(node, 'y'):
                if getattr(node, 'x', 0) == 0 or node.y == 0:
                    self._drawNode(node)
                
    def _drawNode(self, node):
        """ Allocate node with x=0 and y=0. """

        try:
            parents = node.getParents()
            if not parents:
                print("EMPTY NODE ask JM")
                return
            maxParent = parents[0]

            for p in parents[1:]:
                if p.y > maxParent.y:
                    maxParent = p

            siblings = maxParent.getChilds()

            if len(siblings) == 1:
                node.x = maxParent.x
                node.y = maxParent.y + self.DY
            else:
                rightSibling = siblings[0]
                for s in siblings:
                    if s.x > rightSibling.x:
                        rightSibling = s
                node.x = rightSibling.x + rightSibling.width/2 + self.DX + node.width/2
                node.y = rightSibling.y
        except Exception as e:
            from pyworkflow.utils import envVarOn
            if Config.debugOn():
                print("Can't draw node: %s" % node, e)
                import traceback
                traceback.print_stack()
            else:
                # Do nothing
                return


class LevelTreeLayout(GraphLayout):
    """ Organize the nodes of the graph by levels.
    It will recursively organize childs and then
    fit two sibling trees. """
    
    def __init__(self):
        GraphLayout.__init__(self)
        self.DY = 65
        self.DX = 15
        self.maxLevel = 9999
                        
    def draw(self, graph, **kwargs):
        """ Organize nodes of the graph in the plane.
        Nodes should have: x, y, width and height attributes
        x and y will be modified.
        """
        rootNode = graph.getRoot()
        
        # Setup empty layout for each node
        for node in graph.getNodes():
            node._layout = {}
            
        # Do some level initialization on each node
        self._setLayoutLevel(rootNode,  1, None)
        self._computeNodeOffsets(rootNode, 1)
        # Compute extreme left limit
        m = 9999
        for left, _ in rootNode._layout['hLimits']:
            m = min(m, left)
        
        self._applyNodeOffsets(rootNode, -m + self.DY)
        
        # Clean temporary _layout attributes
        for node in graph.getNodes():
            del node._layout
        
    def _setLayoutLevel(self, node, level, parent):
        """ Iterate over all nodes and set _layout dict.
        Also set the node level, which is defined
        as the max level of a parent + 1
        """
        if level > self.maxLevel:
            return 
        
        layout = node._layout
        
        if level > layout.get('level', 0):
            # Calculate the y-position depending on the level
            # and the delta-Y (DY)
            node.y = level * self.DY
            layout['level'] = level
            layout['parent'] = parent
            half = node.width / 2
            layout['half'] = half
            layout['hLimits'] = [[-half, half]]
            layout['offset'] = 0
    
            if self.__isNodeExpanded(node):
                for child in node.getChilds():
                    self._setLayoutLevel(child, level+1, node)
    
    def __isNodeExpanded(self, node):
        """ Check if the status of the node is expanded or collapsed. """
        return getattr(node, 'expanded', True)
        
    def __setNodeOffset(self, node, offset):
        node._layout['offset'] = offset
        
    def __getNodeHalf(self, node):
        return node._layout['half']
    
    def __getNodeChilds(self, node):
        """ Return the node's childs that have been 
        visited by this node first (its 'parent')
        """
        if self.__isNodeExpanded(node):
            return [c for c in node.getChilds() if c._layout['parent'] is node]
        else:
            return []  # treat collapsed nodes as if they have no childs
        
    def _computeNodeOffsets(self, node, level):
        """ Position a parent node and its childs.
        Only this sub-tree will be considered at this point.
        Then it will be adjusted with node siblings.
        """
        if level > self.maxLevel:
            return
        
        childs = self.__getNodeChilds(node)
        n = len(childs)

        if n > 0:
            for c in childs:
                self._computeNodeOffsets(c, level + 1)
                
            if n > 1:
                offset = 0
                # Keep right limits to compute the separation between siblings
                # some times it not enough to compare with the left sibling
                # for some child levels of the node
                rightLimits = [r for l, r in childs[0]._layout['hLimits']]
                
                for i in range(n-1):
                    sep = self._getChildsSeparation(childs[i], childs[i+1], rightLimits)
                    offset += sep
                    c = childs[i+1]
                    self.__setNodeOffset(c, offset)
                
                half0 = self.__getNodeHalf(childs[0])
                total = half0 + offset + self.__getNodeHalf(childs[-1])
                half = total / 2
                for c in childs:
                    self.__setNodeOffset(c, c._layout['offset'] - half + half0)
            else:
                self.__setNodeOffset(childs[0], 0)
            
            self._computeHLimits(node)      
    
    def _computeHLimits(self, node):
        """ This function will traverse the tree
        from node to build the left and right profiles(hLimits)
        for each level of the tree
        """
        layout = node._layout
        hLimits = layout['hLimits']

        childs = self.__getNodeChilds(node)
        
        for child in childs:
            count = 1
            layout = child._layout
            for l, r in layout['hLimits']:
                l += layout['offset']
                r += layout['offset']

                if count < len(hLimits):
                    if l < hLimits[count][0]:
                        hLimits[count][0] = l
                    if r > hLimits[count][1]:
                        hLimits[count][1] = r
                else:
                    hLimits.append([l, r])
                count += 1
                
    def _getChildsSeparation(self, child1, child2, rightLimits):
        """Calculate separation between siblings
        at each height level"""
        sep = 0 
        hL2 = child2._layout['hLimits']
        n1 = len(rightLimits)
        n2 = len(hL2)
        h = min(n1, n2)
            
        for i in range(h):
            right = rightLimits[i]
            left = hL2[i][0]            
            if left + sep < right:
                sep = right - left
            rightLimits[i] = hL2[i][1]
                
        if n1 > n2:
            # If there are more levels in the rightLimits
            # updated the last ones like if they belong 
            # to next sibling is is now (sep + self.DX) away
            for i in range(h, n1):
                rightLimits[i] -= sep + self.DX
        else:
            # If the current right sibling has more levels
            # just add them to the current rightLimits
            for i in range(h, n2):
                rightLimits.append(hL2[i][1])
                
        return sep + self.DX
    
    def _applyNodeOffsets(self, node, x):
        """ Adjust the x-position of the nodes by applying the offsets.
        """
        if node._layout['level'] == self.maxLevel:
            return 
        
        layout = node._layout
        node.x = x + layout['offset']
        
        childs = self.__getNodeChilds(node)
        
        for child in childs:
            self._applyNodeOffsets(child, node.x)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *
# * [1] SciLifeLab, Stockholm University
# *
# * This program is free software: you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation, either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program.  If not, see <https://www.gnu.org/licenses/>.
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import os
import json
import datetime as dt
from collections import OrderedDict
from configparser import ConfigParser

import pyworkflow as pw
import pyworkflow.object as pwobj
from pyworkflow.mapper import SqliteMapper
from pyworkflow.utils import isModuleAFolder


class ProjectSettings(pwobj.OrderedObject):
    """ Store settings related to a project. """

    COLOR_MODE_STATUS = 0
    COLOR_MODE_LABELS = 1
    COLOR_MODE_AGE = 2
    COLOR_MODES = (COLOR_MODE_STATUS, COLOR_MODE_LABELS, COLOR_MODE_AGE)

    def __init__(self, confs={}, **kwargs):
        pwobj.OrderedObject.__init__(self, **kwargs)
        self.config = ProjectConfig()
        # Store the current view selected by the user
        self.currentProtocolsView = pwobj.String()
        # Store the color mode: 0= Status, 1=Labels, ...
        self.colorMode = pwobj.Integer(ProjectSettings.COLOR_MODE_STATUS)
        # Store graph nodes positions and other info
        self.nodeList = NodeConfigList()
        self.labelsList = LabelsList()  # Label list
        self.mapper = None  # This should be set when load, or write
        self.runsView = pwobj.Integer(1)  # by default the graph view
        self.readOnly = pwobj.Boolean(False)
        self.runSelection = pwobj.CsvList(int)  # Store selected runs
        self.dataSelection = pwobj.CsvList(int)  # Store selected runs
        # Some extra settings stored, now mainly used
        # from the webtools
        # Time when the project was created
        self.creationTime = pwobj.String(dt.datetime.now())
        # Number of days that this project is active
        # if None, the project will not expire
        # This is used in webtools where a limited time
        # is allowed for each project
        self.lifeTime = pwobj.Integer()
        # Set a disk quota for the project (in Gb)
        # if None, quota is unlimited
        self.diskQuota = pwobj.Integer()

    def commit(self):
        """ Commit changes made. """
        self.mapper.commit()

    def getRunsView(self):
        return self.runsView.get()

    def setRunsView(self, value):
        self.runsView.set(value)

    def getReadOnly(self):
        return self.readOnly.get()

    def setReadOnly(self, value):
        self.readOnly.set(value)

    def getCreationTime(self):
        return self.creationTime.datetime()

    def setCreationTime(self, value):
        self.creationTime.set(value)

    def getLifeTime(self):
        return self.lifeTime.get()

    def setLifeTime(self, value):
        self.lifeTime.set(value)

    def getConfig(self):
        return self.config

    def getProtocolView(self):
        return self.currentProtocolsView.get()

    def setProtocolView(self, protocolView):
        """ Set the new protocol Menu given its index.
        The new ProtocolMenu will be returned.
        """
        self.currentProtocolsView.set(protocolView)

    def getColorMode(self):
        return self.colorMode.get()

    def setColorMode(self, colorMode):
        """ Set the color mode to use when drawing the graph.
        """
        self.colorMode.set(colorMode)

    def statusColorMode(self):
        return self.getColorMode() == self.COLOR_MODE_STATUS

    def labelsColorMode(self):
        return self.getColorMode() == self.COLOR_MODE_LABELS

    def ageColorMode(self):
        return self.getColorMode() == self.COLOR_MODE_AGE

    def write(self, dbPath=None):
        self.setName('ProjectSettings')
        if dbPath is not None:
            self.mapper = SqliteMapper(dbPath, globals())
        else:
            if self.mapper is None:
                raise Exception("Can't write ProjectSettings without "
                                "mapper or dbPath")

        self.mapper.deleteAll()
        self.mapper.insert(self)
        self.mapper.commit()

    def getNodes(self):
        return self.nodeList

    def getNodeById(self, nodeId):
        return self.nodeList.getNode(nodeId)

    def addNode(self, nodeId, **kwargs):
        return self.nodeList.addNode(nodeId, **kwargs)

    def getLabels(self):
        return self.labelsList

    @classmethod
    def load(cls, dbPath):
        """ Load a ProjectSettings from dbPath. """
        classDict = dict(globals())
        classDict.update(pwobj.__dict__)
        mapper = SqliteMapper(dbPath, classDict)
        settingList = mapper.selectByClass('ProjectSettings')
        n = len(settingList)

        if n == 0:
            raise Exception("Can't load ProjectSettings from %s" % dbPath)
        elif n > 1:
            raise Exception("Only one ProjectSettings is expected in db, "
                            "found %d in %s" % (n, dbPath))

        settings = settingList[0]
        settings.mapper = mapper

        return settings


class ProjectConfig(pwobj.OrderedObject):
    """A simple base class to store ordered parameters"""

    def __init__(self, **args):
        pwobj.OrderedObject.__init__(self, **args)
        self.logo = pwobj.String('scipion_logo_small.gif')
        # Do not store this object, unless we implement some kind of
        # icon customization
        self._objDoStore = False


class MenuConfig(object):
    """Menu configuration in a tree fashion.
    Each menu can contains submenus.
    Leaf elements can contain actions"""

    def __init__(self, text=None, value=None,
                 icon=None, tag=None, **kwargs):
        """Constructor for the Menu config item.
        Arguments:
          text: text to be displayed
          value: internal value associated with the item.
          icon: display an icon with the item
          tag: put some tags to items
        **args: pass other options to base class.
        """
        self.text = pwobj.String(text)
        self.value = pwobj.String(value)
        self.icon = pwobj.String(icon)
        self.tag = pwobj.String(tag)
        self.shortCut = pwobj.String(kwargs.get('shortCut', None))
        self.childs = pwobj.List()
        self.openItem = pwobj.Boolean(kwargs.get('openItem', False))

    def addSubMenu(self, text, value=None, **args):
        subMenu = type(self)(text, value, **args)
        self.childs.append(subMenu)
        return subMenu

    def __iter__(self):
        for v in self.childs:
            yield v

    def __len__(self):
        return len(self.childs)

    def isEmpty(self):
        return len(self.childs) == 0


class NodeConfig(pwobj.Scalar):
    """ Store Graph node information such as x, y. """

    def __init__(self, nodeId=0, x=None, y=None, selected=False, expanded=True):
        pwobj.Scalar.__init__(self)
        # Special node id 0 for project node
        self._values = {'id': nodeId,
                        'x': pwobj.Integer(x).get(0),
                        'y': pwobj.Integer(y).get(0),
                        'selected': selected,
                        'expanded': expanded,
                        'labels': []}

    def _convertValue(self, value):
        """Value should be a str with comma separated values
        or a list.
        """
        self._values = json.loads(value)

    def getObjValue(self):
        self._objValue = json.dumps(self._values)
        return self._objValue

    def get(self):
        return self.getObjValue()

    def getId(self):
        return self._values['id']

    def setX(self, x):
        self._values['x'] = x

    def getX(self):
        return self._values['x']

    def setY(self, y):
        self._values['y'] = y

    def getY(self):
        return self._values['y']

    def setPosition(self, x, y):
        self.setX(x)
        self.setY(y)

    def getPosition(self):
        return self.getX(), self.getY()

    def setSelected(self, selected):
        self._values['selected'] = selected

    def isSelected(self):
        return self._values['selected']

    def setExpanded(self, expanded):
        self._values['expanded'] = expanded

    def isExpanded(self):
        return self._values['expanded']

    def setLabels(self, labels):
        self._values['labels'] = labels

    def getLabels(self):
        return self._values.get('labels', None)

    def __str__(self):
        return 'NodeConfig: %s' % self._values


class NodeConfigList(pwobj.List):
    """ Store all nodes information items and 
    also store a dictionary for quick access
    to nodes query.
    """

    def __init__(self):
        self._nodesDict = {}
        pwobj.List.__init__(self)

    def getNode(self, nodeId):
        return self._nodesDict.get(nodeId, None)

    def addNode(self, nodeId, **kwargs):
        node = NodeConfig(nodeId, **kwargs)
        self._nodesDict[node.getId()] = node
        self.append(node)
        return node

    def updateDict(self):
        self._nodesDict.clear()
        for node in self:
            self._nodesDict[node.getId()] = node

    def clear(self):
        pwobj.List.clear(self)
        self._nodesDict.clear()


class Label(pwobj.Scalar):
    """ Store Label information """

    def __init__(self, labelId=None, name='', color=None):
        pwobj.Scalar.__init__(self)
        # Special node id 0 for project node
        self._values = {'id': labelId,
                        'name': name,
                        'color': color}

    def _convertValue(self, value):
        """Value should be a str with comma separated values
        or a list.
        """
        self._values = json.loads(value)

    def getObjValue(self):
        self._objValue = json.dumps(self._values)
        return self._objValue

    def get(self):
        return self.getObjValue()

    def getId(self):
        return self._values['id']

    def getName(self):
        return self._values['name']

    def setName(self, newName):
        self._values['name'] = newName

    def setColor(self, color):
        self._values['color'] = color

    def getColor(self):
        return self._values.get('color', None)

    def __str__(self):
        return 'Label: %s' % self._values

    def __eq__(self, other):
        return self.getName() == other.getName()


class LabelsList(pwobj.List):
    """ Store all labels information"""

    def __init__(self):
        self._labelsDict = {}
        pwobj.List.__init__(self)

    def getLabel(self, name):
        return self._labelsDict.get(name, None)

    def addLabel(self, label):
        self._labelsDict[label.getName()] = label
        self.append(label)
        return label

    def updateDict(self):
        self._labelsDict.clear()
        for label in self:
            self._labelsDict[label.getName()] = label

    def deleteLabel(self, label):
        self._labelsDict.pop(label.getName())
        self.remove(label)

    def clear(self):
        pwobj.List.clear(self)
        self._labelDict.clear()



#!/usr/bin/env python
# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *
# * [1] SciLifeLab, Stockholm University
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import sys
import os
from datetime import datetime

import pyworkflow as pw
import pyworkflow.utils as pwutils
from pyworkflow.project import Manager
from pyworkflow.gui.project import ProjectWindow


def usage(error):
    print("""
    ERROR: %s
    
    Usage: scipion python scripts/load_project.py /path/to/project
        This script allows to quickly load a project folder without
        importing that in the general user data folder
    """ % error)
    sys.exit(1)    


n = len(sys.argv)

if n < 2 or n > 3:
    usage("Incorrect number of input parameters")
    
pathToProj = os.path.abspath(sys.argv[1])

now = datetime.now()
tempSpace = "loading-%s" % now.strftime('%Y%m%d-%H%M%S')
customUserData = os.path.join(pw.Config.SCIPION_USER_DATA, 'tmp', tempSpace)

projectsDir = os.path.join(customUserData, 'projects')
pwutils.makePath(projectsDir)

print("Loading projects from:\n", projectsDir)

projName = os.path.basename(pathToProj)
pwutils.createAbsLink(pathToProj, os.path.join(projectsDir, projName))
 
# Create a new project
manager = Manager(workspace=customUserData)

proj = manager.loadProject(projName)
projPath = manager.getProjectPath(projName)


class EditorProjectWindow(ProjectWindow):
    def close(self, e=None):
        try:
            print("Deleting temporary folder: ", customUserData)
            pwutils.cleanPath(customUserData)
        except Exception as ex:
            print("Error: ", ex)
        ProjectWindow.close(self, e)


projWindow = EditorProjectWindow(projPath)
projWindow.show()



#!/usr/bin/env python
# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
#                Laura del Cano         (ldelcano@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

from .tests import *

try:
    from unittest.runner import _WritelnDecorator  # Python 2.7+
except ImportError:
    from unittest import _WritelnDecorator  # Python <2.6

DataSet(name='igbmc_gempicker', folder='igbmc_gempicker',
        files={
            'allMics': 'micrographs/*.mrc',
            'mic1': 'micrographs/KLH_Dataset_I_Test_0001.mrc',
            'mic2': 'micrographs/KLH_Dataset_I_Test_0002.mrc',
            'mask1': 'masks/mask_sideview.tif',
            'mask2': 'masks/mask_topview.tif',
            'templates': 'templates/*.mrc'})

DataSet(name='xmipp_tutorial', folder='xmipp_tutorial',
        files={
            'aligned_particles': 'gold/aligned_particles.sqlite',
            'allMics': 'micrographs/*.mrc',
            'boxingDir': 'pickingEman/info',
            'boxingFile': 'pickingEman/info/BPV_1386_info.json',
            'coordsGoldSqlite': 'gold/coordinates_gold.sqlite',
            'ctfGold': 'gold/xmipp_ctf.ctfparam',
            'ctf': 'ctfs/ctfs.sqlite',
            'images10': 'gold/images10.xmd',
            'mic1': 'micrographs/BPV_1386.mrc',
            'mic2': 'micrographs/BPV_1387.mrc',
            'mic3': 'micrographs/BPV_1388.mrc',
            'micsGoldSqlite': 'gold/micrographs_gold.sqlite',
            'micsGoldSqlite2': 'gold/micrographs2_gold.sqlite',
            'micsGoldXmd': 'gold/micrographs_gold.xmd',
            'micsSqlite': 'micrographs/micrographs.sqlite',
            'particles': 'particles/*.hdf',
            'particles1': 'particles/BPV_1386_ptcls.hdf',
            'particles2': 'particles/BPV_particles.sqlite',
            'particles3': 'particles/BPV_particles_aligned.sqlite',
            'posAllDir': 'pickingXmipp/pickedAll',
            'posSupervisedDir': 'pickingXmipp/pickedSupervised',
            'rctCoords': 'rct/pickingXmipp',
            'rctMicsT': 'rct/micrographs/*T.mrc',
            'rctMicsU': 'rct/micrographs/*U.mrc',
            'vol1': 'volumes/BPV_scale_filtered_windowed_64.vol',
            'vol2': 'volumes/volume_1_iter_002.mrc',
            'vol3': 'volumes/volume_2_iter_002.mrc',
            'vol4': 'volumes/helix_59_4__6_7.vol',
            'vol110': 'volumes/BPV_scale_filtered_windowed_110.vol',
            'volumes': 'volumes/*.mrc',
            'mask2d': 'masks/mask2d.xmp',
            'mask3d': 'masks/mask.vol',
            'ctfsDir': 'ctfs'})

DataSet(name='mda', folder='hemoglobin_mda',
        files={
            'particles': 'particles/*.spi',
            'particlesDir': 'particles',
            'volumes': 'volumes/*.spi',
            'volumesDir': 'volumes',
            'averages': 'averages/averages.stk',
            'averagesDir': 'averages'})

DataSet(name='nma', folder='nma',
        files={
            'pdb': 'pdb/AK.pdb',
            'vol': 'volumes/AK_LP10.vol',
            'particles': 'particles/img.stk'})

DataSet(name='powerfit', folder='powerfit',
        files={
            'pdb': '3j79-pdb-bundle1.pdb',
            'vol': 'ribosome_4.vol'})

DataSet(name='tomo', folder='xmipp_tomo_test',
        files={
            'vol1': 'volumes/subvols_6E6.001.mrc.spi',
            'vol2': 'volumes/subvols_6E6.002.mrc.spi',
            'vol3': 'volumes/subvols_6E6.003.mrc.spi',
            'volumes': 'volumes/*.spi'})

DataSet(name='relion_tutorial', folder='relion_tutorial',
        files={
            'allMics': 'micrographs/*.mrc',
            'boxingDir': 'pickingEman',
            'input_particles': 'gold/input_particles.star',
            'particles': 'gold/particles.sqlite',
            'posAllDir': 'pickingXmipp',
            'relion_it020_data': 'gold/relion_it020_data.star',
            'relion_it020_data_newlabels': 'gold/relion_it020_data_newlabels.star',
            'volume': 'volumes/reference.mrc',
            'import1_data_star': 'import/case1/classify3d_small_it038_data.star',
            'import2_data_star': 'import/case2/relion_it015_data.star',
            'particles_binary': '006.stk'
        })

DataSet(name='relion30_tutorial', folder='relion30_tutorial', files={})

DataSet(name='ribo_movies', folder='ribo_movies',
        files={
            'movies': 'movies/1??_*.mrcs',
            'posAllDir': 'pickingXmipp',
            'volume': 'volumes/reference.mrc'})

DataSet(name='model', folder='model',
        files={
            'classesSelection': 'gold/classes_selection.sqlite',
            'modelGoldSqlite': 'gold/model_gold.sqlite',
            'modelGoldXml': 'gold/model_gold.xml'})

DataSet(name='rct', folder='rct',
        files={
            'classes': 'classes/classes2D_stable_core.sqlite',
            'positions': 'positions',
            'tilted': 'micrographs/F_rct_t*.tif',
            'untilted': 'micrographs/F_rct_u*.tif'})

DataSet(name='groel', folder='groel',
        files={
            'averages': 'classes/level_classes.stk',
            'mic1': 'micrographs/groel-stain.mrc',
            'pickingDogpicker': 'pickingDogpicker'})

# TODO: Files to be deleted after checking no plugin or old scipion
#  installations (1.2, ...) are using them:
#  reconstRotandShiftFlip_Gold_output.vol (moved to reconstruction/gold)
#  reconstRotandShift_Gold_output.vol (moved to reconstruction/gold)
#  reconstRotOnly_Gold_output.vol (moved to reconstruction/gold)

DataSet(name='emx', folder='emx',
        files={
            'alignFlip': 'alignment/alignFlip.mrcs',
            'alignFlip2': 'alignment/alignFlip2.mrcs',
            'alignShiftRotExp': 'alignment/Test1/images.mrc',
            'alignShiftRotExpGold': 'alignment/gold/align_shift_rot_exp_gold.mrcs',

            'alignReconst': 'alignment/recons.mrcs',
            'alignReconstShift': 'alignment/recons_sh.mrcs',
            'alignRotOnly': 'alignment/align_rot_only.mrcs',
            'alignRotOnlyGold': 'alignment/gold/align_rot_only_gold.mrcs',
            # TODO: transform geometry not working in stacks of volumes
            'alignRotOnly3D': 'alignment/align_rot_only_3d.mrcs',
            # 'alignRotShift': 'alignment/align_rot_shift.mrcs',
            'alignShiftOnly': 'alignment/align_shift_only.mrcs',
            'alignShiftOnlyGold':
                'alignment/gold/align_shift_only_gold.mrcs',
            # TODO: transform geometry not working in stacks of volumes
            'alignShiftOnly3D': 'alignment/align_shift_only_3d.mrcs',
            'alignShiftRot': 'alignment/align_shift_rot.mrcs',
            'alignShiftRotGold': 'alignment/gold/align_shift_rot_gold.mrcs',
            # TODO: transform geometry not working in stacks of volumes
            'alignShiftRot3D': 'alignment/align_shift_rot_3d.mrcs',
            'coordinatesGoldT1':
                'coordinates/Test1/coordinates_gold.sqlite',
            'coordinatesT1': 'coordinates/Test1/coordinates.emx',
            'defocusParticleT2': 'defocusParticle/particles.emx',
            'emxMicrographCtf1': 'MicrographsCTF/ctfindCTFEstimation.emx',
            'emxMicrographCtf1Gold':
                'MicrographsCTF/ctfindCTFEstimation.sqlite',
            'micrographsGoldT2': 'defocusParticle/micrographs_gold.sqlite',
            'particlesGoldT2': 'defocusParticle/particles_gold.sqlite',
            'reconstRotOnly': 'reconstruction/reconstRotOnly.mrcs',
            'reconstRotOnlyGold':
                'reconstruction/gold/reconstRotOnly_output.vol',
            'reconstRotandShift': 'reconstruction/reconstRotandShift.mrcs',
            'reconstRotandShiftGold':
                'reconstruction/gold/reconstRotandShift_output.vol',
            'reconstRotandShiftFlip':
                'reconstruction/reconstRotandShiftFlip.mrcs',
            'reconstRotandShiftFlipGold':
                'reconstruction/gold/reconstRotandShiftFlip_output.vol',
            'reconstRotOnlyGoldRln':
                'reconstruction/gold/reconstRotOnly_rln_output.vol',
            'reconstRotandShiftGoldRln':
                'reconstruction/gold/reconstRotandShift_rln_output.vol',
            'reconstRotandShiftFlipGoldRln':
                'reconstruction/gold/reconstRotandShiftFlip_rln_output.vol'
        })

DataSet(name='CTFDiscrepancy', folder='CTFDiscrepancy',
        files={
            'emxMicrographCtf1': 'ctfindCTFEstimation.emx',
            'emxMicrographCtf2': 'xmipp3CTFEstimation.emx',
            'emxMicrographCtf3': 'xmipp3CTFEstimation2.emx',
            'ctfsGold': 'ctfs.sqlite'})

DataSet(name='movies', folder='movies',
        files={'movie1': 'ribo/Falcon_2012_06_12-14_33_35_0_movie.mrcs',
               'movie2': 'ribo/Falcon_2012_06_12-16_55_40_0_movie.mrcs',
               'movie3': 'ribo/Falcon_2012_06_12-17_26_54_0_movie.mrcs',
               'movie4': 'qbeta/qbeta.mrc',
               'movie5': 'cct/cct_1.em',
               'falcon2012_1': 'Falcon_2012_06_12-16_55_40_0_movie.mrcs',
               'mrcGain': 'gain.mrc',
               'mrcDark': 'dark.mrc',
               'spiGain': 'gain.spi',
               'spiDark': 'dark.spi',
               })
DataSet(name='riboMovies', folder='riboMovies',
        files={'movie1': '30_movie_gc_window.mrcs',
               'movie2': '35_movie_gc_window.mrcs',
               'movie3': '130_movie_gc_window.mrcs',
               'movie4': '165_movie_gc_window.mrcs',
               'movie5': '270_movie_gc_window.mrcs',
               'allMovies': '*.mrcs'})

DataSet(name='jmbFalconMovies', folder='jmbFalconMovies',
        files={'movie1': 'Falcon_2014_07_16-16_32_43_0.mrcs',
               'movie2': 'Falcon_2014_07_16-21_37_37_0.mrcs',
               'movie3': 'Falcon_2014_07_16-23_30_50_0.mrcs',
               'movie4': 'Falcon_2014_07_17-03_21_57_0.mrcs',
               'movie5': 'Falcon_2014_07_17-06_55_56_0.mrcs',
               'allMovies': '*.mrcs'})

DataSet(name='grigorieff', folder='grigorieff',
        files={'particles_rot_shift': 'particles.sqlite',
               'ref_volume': 'ref_volume.vol',
               'ctfsDir': 'ctfs'})

DataSet(name='initial_volume', folder='initial_volume',
        files={'groel': 'groel_64.stk',
               'ribosome': 'ribosomeFrank_64_preprocessed.stk',
               'bpv': 'bpv_100.stk'})

DataSet(name='general', folder='general',
        files={'vol_helix': 'volumes/helix_59_4__6_7.vol'})

DataSet(name='SymVirus', folder='SymVirus',
        files={'whole_vol_half': 'whole_vol_half.vol',
               'large_vol_half_th': 'large_vol_half_th.vol',
               'small_vol_half_th': 'small_vol_half_th.vol'})

DataSet(name='xmipp_programs', folder='xmipp_programs', files={})

DataSet(name='resmap', folder='resmap',
        files={'betagal': 'betaGal.mrc',
               'betagal_half1': 'betaGal_half01.mrc',
               'betagal_half2': 'betaGal_half02.mrc',
               'betagal_mask': 'betaGal_mask.mrc'})

DataSet(name='10010', folder='10010',
        files={'initialVolume': 'initialVolume.vol',
               'particles': 'particles.sqlite'})

DataSet(name='model_building_tutorial', folder='model_building_tutorial',
        files={'vol1': 'volumes/1ake_4-5A.mrc',
               'pdb1': 'PDBs/1ake_start.pdb'})

DataSet(name='eman', folder='eman',
        files={
            'micU': 'mics/ip3r10252011-0005_0-2.hdf',
            'micT': 'mics/ip3r10252011-0005_10.hdf',
            'volume': '3DmapIP3R1_clip_376apix.mrc',
            'coords': 'coords/*.json',
            'ctfs': 'ctfs/'
        })



#!/usr/bin/env python
# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *
# * [1] SciLifeLab, Stockholm University
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

import os
import sys
import time
import argparse

from pyworkflow.protocol import (getProtocolFromDb,
                                 STATUS_FINISHED, STATUS_ABORTED, STATUS_FAILED,
                                 Set, Protocol)

# Add callback for remote debugging if available.
from pyworkflow.utils import prettyTimestamp, getFileLastModificationDate

try:
    from rpdb2 import start_embedded_debugger
    from signal import signal, SIGUSR2

    signal(SIGUSR2, lambda sig, frame: start_embedded_debugger('a'))
except ImportError:
    pass


class RunScheduler:
    """ Check that all dependencies are met before launching a run. """

    def _parseArgs(self):
        parser = argparse.ArgumentParser()
        _addArg = parser.add_argument  # short notation

        _addArg("projPath", metavar='PROJECT_NAME',
                help="Project database path.")

        _addArg("dbPath", metavar='DATABASE_PATH',
                help="Protocol database path.")

        _addArg("protId", type=int, metavar='PROTOCOL_ID',
                help="Protocol ID.")

        _addArg("--sleep_time", type=int, default=15,
                dest='sleepTime', metavar='SECONDS',
                help="Sleeping time (in seconds) between updates.")

        _addArg("--wait_for", nargs='*', type=int, default=[],
                dest='waitProtIds', metavar='PROTOCOL_ID',
                help="List of protocol ids that should be not running "
                     "(i.e, finished, aborted or failed) before this "
                     "run will be executed.")

        self._args = parser.parse_args()

    def _loadProtocol(self):
        return getProtocolFromDb(self._args.projPath,
                                 self._args.dbPath,
                                 self._args.protId, chdir=True)

    def main(self):
        self._parseArgs()

        stopStatuses = [STATUS_FINISHED, STATUS_ABORTED, STATUS_FAILED]

        # Enter to the project directory and load protocol from db
        protocol = self._loadProtocol()
        mapper = protocol.getMapper()

        log = open(protocol.getScheduleLog(), 'w')
        pid = os.getpid()
        protocol.setPid(pid)

        prerequisites = list(map(int, protocol.getPrerequisites()))

        def _log(msg):
            log.write("%s: %s\n" % (prettyTimestamp(), msg))
            log.flush()

        _log("Scheduling protocol %s, pid: %s, prerequisites: %s"
             % (protocol.getObjId(), pid, prerequisites))

        mapper.store(protocol)
        mapper.commit()
        mapper.close()

        # Keep track of the last time the protocol was checked and
        # its modification date to avoid unnecessary db opening
        lastCheckedDict = {}
        updatedProtocols = []

        def _updateProtocol(protocol, project):

            protId = protocol.getObjId()

            if protId in updatedProtocols:
                return

            protDb = protocol.getDbPath()

            if os.path.exists(protDb):
                lastChecked = lastCheckedDict.get(protId, None)
                lastModified = getFileLastModificationDate(protDb)

                if lastChecked is None or (lastModified > lastChecked):
                    project._updateProtocol(protocol,
                                            skipUpdatedProtocols=False)
                    _log("Updated protocol %s" % protId)
                    updatedProtocols.append(protId)

        def _getProtocolFromPointer(pointer):
            """
            The function return a protocol from an attribute

               A) When the pointer points to a protocol

               B) When the pointer points to another object (INDIRECTLY).
                  - The pointer has an _extended value (new parameters
                    configuration in the protocol)

               C) When the pointer points to another object (DIRECTLY).
                  - The pointer has not an _extended value (old parameters
                    configuration in the protocol)
            """
            output = pointer.get()
            if isinstance(output, Protocol):  # case A
                protocol = output
            else:
                if pointer.hasExtended():  # case B
                    protocol = pointer.getObjValue()
                else:  # case C
                    protocol = self.getProject().getProtocol(
                        output.getObjParentId())
            return protocol

        while True:
            protocol = self._loadProtocol()
            project = protocol.getProject()

            # Check if there are missing inputs
            missing = False
            # Check if there are input protocols failed or aborted
            failedInputProtocols = False

            # Clear the list of protocols updated in the previous loop
            updatedProtocols.clear()

            _log("Checking input data...")
            # FIXME: This does not cover all the cases:
            # When the user registers new coordinates after clicking the
            # "Analyze result" button, this action is registered in the project.sqlite
            # and not in it's own run.db and never gets updated. It is not critical and
            # will only affect a combination of json import with extended that will
            # appear after clicking on the "Analyze result" button.
            for key, attr in protocol.iterInputAttributes():
                inputProt = _getProtocolFromPointer(attr)
                _updateProtocol(inputProt, project)
                if (inputProt.getStatus() == STATUS_ABORTED or
                        inputProt.getStatus() == STATUS_FAILED):
                    failedInputProtocols = True

            if not failedInputProtocols:
                _updateProtocol(protocol, project)
                validation = protocol.validate()
                if len(validation) > 0:
                    missing  = True
                    _log("%s doesn't validate:\n\t- %s"
                         % (protocol.getObjLabel(),
                            '\n\t- '.join(validation)))
                elif not protocol.worksInStreaming():
                    for key, attr in protocol.iterInputAttributes():
                        inSet = attr.get()
                        if isinstance(inSet, Set) and inSet.isStreamOpen():
                            missing = True
                            _log("Waiting for closing %s... "
                                 "(%s does not work in streaming)"
                                 % (inSet, protocol))
                            break

                if not missing:
                    inputProtocolDict = protocol.inputProtocolDict()
                    for prot in inputProtocolDict.values():
                        _updateProtocol(prot, project)

            _log("Checking prerequisites... %s" % prerequisites)

            wait = False  # Check if we need to wait for required protocols
            for protId in prerequisites:
                prot = project.getProtocol(protId)
                if prot is not None:
                    _updateProtocol(prot, project)
                    if prot.getStatus() not in stopStatuses:
                        wait = True
                        _log("   ...waiting for %s" % prot)

            if not missing and not wait:
                break

            project.mapper.commit()
            project.mapper.close()

            _log("Still not ready, sleeping %s seconds...\n"
                 % self._args.sleepTime)
            time.sleep(self._args.sleepTime)

        _log("Launching the protocol >>>>")
        log.close()
        project.launchProtocol(protocol, scheduled=True, force=True)


if __name__ == '__main__':
    try:
        scheduler = RunScheduler()
        scheduler.main()
    except Exception as ex:
        print(ex)
        print("Schedule fail with this parameters: ", sys.argv)



# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *              Jose Gutierrez (jose.gutierrez@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
"""
This module is mainly for the Viewer class, which 
serve as base for implementing visualization tools(Viewer sub-classes).
"""

DESKTOP_TKINTER = 'tkinter'
WEB_DJANGO = 'django'


class Wizard(object):
    """ This is a special case of GUI to help the user
    selecting important parameters.
    The _targets will serve to define to which Definition and 
    parameters the Wizard is defined, it will be a list of tuples such as:
    _targets = [(DefImportMicrographs, ['voltage', sphericalAberration']),
                (DefCTFMicrographs, ['lowRes', 'highRes'])]
    The _environments will serve to define when this wizard can be used.
    For example>
    _environments = [DESKTOP_TKINTER, WEB_DJANGO]
    """
    _targets = []
    _environments = [DESKTOP_TKINTER]
    
    def show(self, form, *params):
        """ This will show up the wizard to select parameters.
        Params:
            form: the protocol form, given access to to all parameters.
                Some times the same wizard will modify several elements
                in the form.
            *params: a list of params to modify, sometimes the wizard can 
                be generic and can be used for different parameters in the
                same form.
        """
        pass
    
    def getView(self):
        """ This method should return the string value of the view in web
        that will respond to this wizard. This method only should be implemented
        in those wizards that have WEB_DJANGO environment defined. 
        """
        return None



#!/usr/bin/env python
# **************************************************************************
# *
# * Authors:     Antonio Poza (Apr 30, 2013)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
import os
import sys
import logging
import logging.config


from pyworkflow.utils import makeFilePath


def getLogConfiguration():
    from pyworkflow import Config
    # Log configuration
    config = {
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'standard': {
                'format': '%(asctime)s %(levelname)s:  %(message)s'
                # TODO: use formattime to show the time less verbose
            },
            'fileFormat': {
                'format': '%(asctime)s %(levelname)s:  %(message)s'
            },
        },
        'handlers': {
            'fileHandler': {
                'level': 'NOTSET',
                'class': 'logging.handlers.RotatingFileHandler',
                'formatter': 'standard',
                'filename': Config.SCIPION_LOG,
                'maxBytes': 100000,
            },
            'consoleHandler': {
                'level': 'NOTSET',
                'class': 'logging.StreamHandler',
                'formatter': 'standard',
            },
        },
        'loggers': {
            '': {
                'handlers': ['consoleHandler', 'fileHandler'],
                'level': 'INFO',
                'propagate': False,
                'qualname': 'pyworkflow',
            },
        }
    }

    # Create the log folder
    os.makedirs(Config.SCIPION_LOGS, exist_ok=True)

    logging.config.dictConfig(config)

    return config


class ScipionLogger:
    def __init__(self, filePath=''):
        """ If filePath is empty string, the general logger is used. """
        self._filePath = filePath
        makeFilePath(self._filePath)

        self.config = getLogConfiguration()

        if self._filePath not in self.config['loggers']:
            self.config['handlers'][self._filePath] = {
                'level': 'NOTSET',
                'class': 'logging.handlers.RotatingFileHandler',
                'formatter': 'fileFormat',
                'filename': self._filePath,
                'maxBytes': 100000}

            self.config['loggers'][self._filePath] = {
                'handlers': [self._filePath],
                'level': 'NOTSET',
                'propagate': False}
            # Note: if we want to see in the console what we also have in
            # run.log, add 'consoleHandler' to the list of 'handlers'.

            logging.config.dictConfig(self.config)
            
        self._log = logging.getLogger(self._filePath) 
        
    def getLog(self):
        return self._log  
    
    def getLogString(self):
        return open(self._filePath, 'r').readlines()  
        
    def info(self, message, redirectStandard=False, *args, **kwargs):
        if redirectStandard:
            print(message)
            sys.stdout.flush()
        self._log.info(message, *args, **kwargs)

    def warning(self, message, redirectStandard=False, *args, **kwargs):
        if redirectStandard:
            print(message)
            sys.stdout.flush()
        self._log.warning(message, *args, **kwargs)
        
    def error(self, message, redirectStandard=False, *args, **kwargs):
        if redirectStandard:
            sys.stderr.write(message + '\n')
            sys.stderr.flush()
        self._log.error(message, *args, **kwargs)    
        
    def close(self):
        if self._filePath in self.config['loggers']:
            del self.config['handlers'][self._filePath]
            del self.config['loggers'][self._filePath]



# **************************************************************************
# *
# * Authors:     Pablo Conesa (pconesa@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
""" Module to test json exporting options"""
# from pyworkflow.em import Pointer, ProtMultiPointerTest
# from pyworkflow.object import PointerList, String
# from tests import *
#
# MY_OUTPUT = "myOutput"

# TODO: Check if some tests from pwem/protocol/protocol_tests can be moved here
# class TestProtocolExport(BaseTest):
#
#     @classmethod
#     def setUpClass(cls):
#         setupTestProject(cls)
#
#     def test_exportMultiPointerToSets(self):
#         """Test how multipointers to are exported and imported"""
#
#         protMpOutput = self.newProtocol(ProtMultiPointerTest,
#                                 objLabel='multipointer for output')
#
#         protMpOutput._defineOutputs(**{MY_OUTPUT: String("hola!")})
#         self.saveProtocol(protMpOutput)
#
#         # Add multiPointers with extended
#         plWithExtended = PointerList()
#         plWithExtended.append(Pointer(protMpOutput, extended=MY_OUTPUT))
#
#         protMp = self.newProtocol(ProtMultiPointerTest,
#                                   objLabel='multipointer with extended export to json',
#                                   mpToAttr=plWithExtended)
#
#         self.saveProtocol(protMp)
#         # Trigger the refresh of the runsGraph!!
#         self.proj._runsGraph = None
#
#         protDict = self.proj.getProtocolsDict()
#
#         # Get the multipointer params items for the second prot
#         # Get the second prot
#         # This has the shape of :  (33, OrderedDict([('object.className', 'ProtMultiPointerTest'), ...) ])
#         ndProtAttrs = protDict[protMp.getObjId()]
#
#         # Look for the mpToAttr
#         for key, value in ndProtAttrs.items():
#
#             if key == "mpToAttr":
#                 self.assertEqual(1, len(value),
#                                  "multipointer param items exceeds the "
#                                  "expected number of items")
#                 self.assertEqual("%s.%s" % (protMpOutput.getObjId(), MY_OUTPUT),
#                                  value[0],
#                                  "Multipointer item value %s seems to be wrong."
#                                  % value[0])



import sys
import os
import time
from traceback import format_exception
import unittest
from os.path import join, relpath

import pyworkflow as pw
import pyworkflow.utils as pwutils
from pyworkflow.project import Manager
from pyworkflow.protocol import MODE_RESTART, getProtocolFromDb

SMALL = 'small'
PULL_REQUEST = 'pull'
DAILY = 'daily'
WEEKLY = 'weekly'


# Procedure to check if a test class has an attribute called _labels and if so
# then it checks if the class test matches any of the labels in input label parameter.
def hasLabel(TestClass, labels):
    # Get _labels attributes in class if any.
    classLabels = getattr(TestClass, '_labels', None)

    # Check if no label in test class.    
    return classLabels is not None and any(l in classLabels for l in labels)


class DataSet:
    _datasetDict = {}  # store all created datasets

    def __init__(self, name, folder, files, url=None):
        """ 
        Params:
            
        #filesDict is dict with key, value pairs for each file
        """
        self._datasetDict[name] = self
        self.folder = folder
        self.path = join(pw.Config.SCIPION_TESTS, folder)
        self.filesDict = files
        self.url = url

    def getFile(self, key):
        if key in self.filesDict:
            return join(self.path, self.filesDict[key])
        return join(self.path, key)

    def getPath(self):
        return self.path

    @classmethod
    def getDataSet(cls, name):
        """
        This method is called every time the dataset want to be retrieved
        """
        assert name in cls._datasetDict, "Dataset: %s dataset doesn't exist." % name

        ds = cls._datasetDict[name]
        folder = ds.folder
        url = '' if ds.url is None else ' -u ' + ds.url

        if not pwutils.envVarOn('SCIPION_TEST_NOSYNC'):
            command = ("%s %s --download %s %s"
                       % (pw.PYTHON, pw.getSyncDataScript(), folder, url))
            print(">>>> %s" % command)
            os.system(command)

        return cls._datasetDict[name]


class BaseTest(unittest.TestCase):
    _labels = [WEEKLY]

    @classmethod
    def getOutputPath(cls, *filenames):
        """Return the path to the SCIPION_HOME/tests/output dir
        joined with filename"""
        return join(cls.outputPath, *filenames)

    @classmethod
    def getRelPath(cls, basedir, filename):
        """Return the path relative to SCIPION_HOME/tests"""
        return relpath(filename, basedir)

    @classmethod
    def launchProtocol(cls, prot, **kwargs):
        """ Launch a given protocol using cls.proj.
        Accepted **kwargs:
            wait: if True the function will return after the protocol runs.
                If not specified, then if waitForOutput is passed, wait is
                false.
            waitForOutputs: a list of expected outputs, ignored if wait=True
        """
        wait = kwargs.get('wait', None)
        waitForOutputs = kwargs.get('waitForOutput', [])

        if wait is None:
            wait = not waitForOutputs

        if getattr(prot, '_run', True):
            cls.proj.launchProtocol(prot, wait=wait)
            if not wait and waitForOutputs:
                while True:
                    time.sleep(10)
                    prot = cls.updateProtocol(prot)
                    if all(prot.hasAttribute(o) for o in waitForOutputs):
                        return prot

        if prot.isFailed():
            print("\n>>> ERROR running protocol %s" % prot.getRunName())
            print("    FAILED with error: %s\n" % prot.getErrorMessage())

            BaseTest.printLastLogLines(prot)

            raise Exception("ERROR launching protocol.")

        if not prot.isFinished() and not prot.useQueue():  # when queued is not finished yet
            print("\n>>> ERROR running protocol %s" % prot.getRunName())

            BaseTest.printLastLogLines(prot)

            raise Exception("ERROR: Protocol not finished")

        return prot

    @staticmethod
    def printLastLogLines(prot):

        lastLines = int(os.environ.get('PROT_LOGS_LAST_LINES', 50))
        print(pwutils.cyanStr("\n*************** LAST %s LINES OF THE LOG *********************\n" % lastLines))
        logLines = prot.getLogsLastLines(lastLines)
        for i in range(0, len(logLines)):
            print(logLines[i])
        print(pwutils.cyanStr("\n*************** END OF THE LOG *********************\n"))

        sys.stdout.flush()

    @classmethod
    def saveProtocol(cls, prot):
        """ Save protocol using cls.proj """
        cls.proj.saveProtocol(prot)

    @classmethod
    def _waitOutput(cls, prot, outputAttributeName):
        """ Wait until the output is being generated by the protocol. """

        def _loadProt():
            # Load the last version of the protocol from its own database
            loadedProt = getProtocolFromDb(prot.getProject().path,
                                           prot.getDbPath(),
                                           prot.getObjId())
            # Close DB connections
            loadedProt.getProject().closeMapper()
            loadedProt.closeMappers()
            return loadedProt

        counter = 1
        prot2 = _loadProt()

        while (not prot2.hasAttribute(outputAttributeName)) and prot2.isActive():
            time.sleep(5)
            prot2 = _loadProt()
            if counter > 1000:
                break
            counter += 1

        # Update the protocol instance to get latest changes
        cls.proj._updateProtocol(prot)

    @classmethod
    def newProtocol(cls, protocolClass, **kwargs):
        """ Create new protocols instances through the project
        and return a newly created protocol of the given class
        """
        # Try to continue from previous execution
        if pwutils.envVarOn('SCIPION_TEST_CONTINUE'):
            candidates = cls.proj.mapper.selectByClass(protocolClass.__name__)
            if candidates:
                c = candidates[0]
                if c.isFinished():
                    setattr(c, '_run', False)
                else:
                    c.runMode.set(MODE_RESTART)
                return c
        return cls.proj.newProtocol(protocolClass, **kwargs)

    @classmethod
    def compareSets(cls, test, set1, set2):
        """ Iterate the elements of both sets and check
        that all elements have equal attributes. """
        for item1, item2 in zip(set1, set2):
            areEqual = item1.equalAttributes(item2)
            if not areEqual:
                print("item 1 and item2 are different: ")
                item1.printAll()
                item2.printAll()
            test.assertTrue(areEqual)

    def assertSetSize(self, setObject, size=None, msg=None, diffDelta=None):
        """ Check if a pyworkflow Set is not None nor is empty, or of a determined size or
        of a determined size with a percentage (base 1) of difference"""
        self.assertIsNotNone(setObject, msg)
        setObjSize = setObject.getSize()

        if size is None:
            # Test is not empty
            self.assertNotEqual(setObjSize, 0, msg)
        else:
            if diffDelta:
                self.assertLessEqual(abs(setObjSize - size), round(diffDelta * size), msg)
            else:
                self.assertEqual(setObjSize, size)

    def assertIsNotEmpty(self, setObject, msg=None):
        """ Check if the pyworkflow object is not None nor is empty"""
        self.assertIsNotNone(setObject, msg)

        self.assertIsNotNone(setObject.get(), msg)


def setupTestOutput(cls):
    """ Create the output folder for a give Test class. """
    cls.outputPath = join(pw.Config.SCIPION_TESTS_OUTPUT, cls.__name__)
    pwutils.cleanPath(cls.outputPath)
    pwutils.makePath(cls.outputPath)


def setupTestProject(cls, writeLocalConfig=False):
    """ Create and setup a Project for a give Test class. """
    projName = cls.__name__
    hostsConfig = None

    if writeLocalConfig:
        hostsConfig = '/tmp/hosts.conf'
        print("Writing local config: %s" % hostsConfig)
        import pyworkflow.protocol as pwprot
        pwprot.HostConfig.writeBasic(hostsConfig)

    if os.environ.get('SCIPION_TEST_CONTINUE', None) == '1':
        proj = Manager().loadProject(projName)
    else:
        proj = Manager().createProject(projName, hostsConf=hostsConfig)

    cls.outputPath = proj.path
    # Create project does not change the working directory anymore
    os.chdir(cls.outputPath)
    cls.projName = projName
    cls.proj = proj


class GTestResult(unittest.TestResult):
    """ Subclass TestResult to output tests results with colors
    (green for success and red for failure)
    and write a report on an .xml file. 
    """
    xml = None
    testFailed = 0
    numberTests = 0

    def __init__(self):
        unittest.TestResult.__init__(self)
        self.startTimeAll = time.time()

    def openXmlReport(self, classname, filename):
        pass

    def doReport(self):
        secs = time.time() - self.startTimeAll
        sys.stderr.write("\n%s run %d tests (%0.3f secs)\n" %
                         (pwutils.greenStr("[==========]"),
                          self.numberTests, secs))
        if self.testFailed:
            sys.stderr.write("%s %d tests\n"
                             % (pwutils.redStr("[  FAILED  ]"),
                                self.testFailed))
        sys.stdout.write("%s %d tests\n"
                         % (pwutils.greenStr("[  PASSED  ]"),
                            self.numberTests - self.testFailed))

    def tic(self):
        self.startTime = time.time()

    def toc(self):
        return time.time() - self.startTime

    def startTest(self, test):
        self.tic()
        self.numberTests += 1

    @staticmethod
    def getTestName(test):
        parts = str(test).split()
        name = parts[0]
        parts = parts[1].split('.')
        classname = parts[-1].replace(")", "")
        return "%s.%s" % (classname, name)

    def addSuccess(self, test):
        secs = self.toc()
        sys.stderr.write("%s %s (%0.3f secs)\n" %
                         (pwutils.greenStr('[ RUN   OK ]'),
                          self.getTestName(test), secs))

    def reportError(self, test, err):
        sys.stderr.write("%s %s\n" % (pwutils.redStr('[   FAILED ]'),
                                      self.getTestName(test)))
        sys.stderr.write("\n%s"
                         % pwutils.redStr("".join(format_exception(*err))))
        self.testFailed += 1

    def addError(self, test, err):
        self.reportError(test, err)

    def addFailure(self, test, err):
        self.reportError(test, err)



# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *
# * [1] SciLifeLab, Stockholm University
# *
# * This program is free software: you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation, either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program.  If not, see <https://www.gnu.org/licenses/>.
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


import re
import collections

from pyworkflow.object import *
from .constants import *


class FormElement(OrderedObject):
    """Base for any element on the form"""
    ATTRIBUTES = ['label', 'expertLevel', 'condition', 'important', 'help',
                  'default', 'paramClass']
    
    def __init__(self, **args):
        OrderedObject.__init__(self, **args)
        self.label = String(args.get('label', None))
        self.expertLevel = Integer(args.get('expertLevel', LEVEL_NORMAL))
        self.condition = String(args.get('condition', None))
        self._isImportant = Boolean(args.get('important', False))
        self.help = String(args.get('help', None))
        # This two list will be filled by the Form
        # which have a global-view of all parameters
        # All param names in which condition appears this param
        self._dependants = []
        # All param names that appears in the condition
        self._conditionParams = []
        
    def isExpert(self):
        return self.expertLevel > LEVEL_NORMAL
    
    def setExpert(self):
        self.expertLevel.set(LEVEL_ADVANCED)
        
    def isImportant(self):
        return self._isImportant.get()
    
    def setImportant(self, value):
        self._isImportant.set(value)
    
    def hasCondition(self):
        return self.condition.hasValue()
    
    def getLabel(self):
        return self.label.get()

    def getHelp(self):
        return self.help.get()
    
    def config(self, **kwargs):
        """ Configure the object and set attributes
        coming in the keyword-arguments, the 
        same as in the __init__
        """
        for key in self.ATTRIBUTES:
            if key in kwargs:
                self.setAttributeValue(key, kwargs.get(key)) 
    
        
class Param(FormElement):
    """Definition of a protocol parameter"""
    def __init__(self, **args):
        FormElement.__init__(self, **args)
        # This should be defined in subclasses
        self.paramClass = args.get('paramClass', None)
        self.default = String(args.get('default', None))

        # Allow pointers (used for scalars)
        self.allowsPointers = args.get('allowsPointers', False)
        self.validators = args.get('validators', [])
        
    def __str__(self):
        return "    label: %s" % self.label.get()
    
    def addValidator(self, validator):
        """ Validators should be callables that 
        receive a value and return a list of errors if so.
        If everything is ok, the result should be an empty list.
        """
        self.validators.append(validator)
        
    def validate(self, value):
        errors = []
        for val in self.validators:
            errors += val(value)
        return errors
    
    def getDefault(self):
        return self.default.get()
    
    def setDefault(self, newDefault):
        self.default.set(newDefault)
    
    
class ElementGroup(FormElement):
    """ Class to group some params in the form.
    Such as: Labeled group or params in the same line. 
    """
    def __init__(self, form=None, **args):
        FormElement.__init__(self, **args)
        self._form = form
        self._paramList = []    
    
    def iterParams(self):
        """ Return key and param for every child param. """
        for name in self._paramList:
            yield name, self._form.getParam(name)

    def addParam(self, paramName, ParamClass, **kwargs):
        """Add a new param to the group"""
        param = ParamClass(**kwargs)
        self._paramList.append(paramName)
        self._form.registerParam(paramName, param)
        return param
    
    def addHidden(self, paramName, ParamClass, **kwargs):
        """Add a hidden parameter to be used in conditions. """
        kwargs.update({'label': '', 'condition': 'False'})
        self.addParam(paramName, ParamClass, **kwargs)

    def addLine(self, lineName, **kwargs):
        
        labelName = lineName
        for symbol in ' ()':
            labelName = labelName.replace(symbol, '_')
        
        return self.addParam(labelName, Line, form=self._form, 
                             label=lineName, **kwargs)        
    
    
# ----------- Some type of ElementGroup --------------------------

class Line(ElementGroup):
    """ Group to put some parameters in the same line. """
    pass


class Group(ElementGroup):
    """ Group some parameters with a labeled frame. """
    pass

    
class Section(ElementGroup):
    """Definition of a section to hold other params"""
    def __init__(self, form, **args):
        ElementGroup.__init__(self, form, **args)
        self.questionParam = String(args.get('questionParam', ''))
    
    def hasQuestion(self):
        """Return True if a question param was set"""
        return self.questionParam.get() in self._paramList
    
    def getQuestionName(self):
        """ Return the name of the question param. """
        return self.questionParam.get()
    
    def getQuestion(self):
        """ Return the question param"""
        return self._form.getParam(self.questionParam.get())

    def addGroup(self, groupName, **kwargs):
        labelName = groupName
        for symbol in ' ()':
            labelName = labelName.replace(symbol, '_')
        
        return self.addParam(labelName, Group, form=self._form, 
                             label=groupName, **kwargs)
        

class Form(object):
    """Store all sections and parameters"""
    def __init__(self, protocol):
        """ Build a Form from a given protocol. """
        object.__init__(self)
        self._sectionList = []  # Store list of sections
        # Dictionary to store all params, grouped by sections
        self._paramsDict = collections.OrderedDict()
        self._lastSection = None
        self._protocol = protocol
        self.addGeneralSection()
        
    def getClass(self):
        return type(self)
        
    def addSection(self, label='', **kwargs):
        """Add a new section"""
        self.lastSection = Section(self, label=label, **kwargs)
        self._sectionList.append(self.lastSection)
        return self.lastSection

    def getSection(self, label):
        """ get section by label from _sectionList"""
        for s in self._sectionList:
            if s.label == label:
                return s
        return

    def addGroup(self, *args, **kwargs):
        return self.lastSection.addGroup(*args, **kwargs)
    
    def addLine(self, *args, **kwargs):
        return self.lastSection.addLine(*args, **kwargs)      

    def registerParam(self, paramName, param):
        """ Register a given param in the form. """
        self._paramsDict[paramName] = param        
        self._analizeCondition(paramName, param)
        
    def addParam(self, *args, **kwargs):
        """Add a new param to last section"""
        return self.lastSection.addParam(*args, **kwargs)
    
    def addHidden(self, *args, **kwargs):
        return self.lastSection.addHidden(*args, **kwargs)
    
    def _analizeCondition(self, paramName, param):
        if param.hasCondition():
            param._conditionParams = []
            tokens = re.split('\W+', param.condition.get())
            for t in tokens:
                if self.hasParam(t):
                    self.getParam(t)._dependants.append(paramName)
                    param._conditionParams.append(t)
                if self._protocol.hasAttribute(t):
                    param._conditionParams.append(t)
                    
    def evalParamCondition(self, paramName):
        """Evaluate if a condition is True for a give param
        with the values of a particular Protocol"""
        param = self.getParam(paramName)

        if not param.hasCondition():
            return True
        condStr = param.condition.get()
        localDict = {}
        globalDict = dict(globals())
        # FIXME: Check why this import is here
        from pyworkflow import Config
        globalDict.update(Config.getDomain().getObjects())

        for t in param._conditionParams:
            if self.hasParam(t) or self._protocol.hasAttribute(t):
                localDict[t] = self._protocol.getAttributeValue(t)

        return eval(condStr, globalDict, localDict)
    
    def validateParams(self, protocol):
        """ Check that all validations of the params in the form
        are met for the protocol param values.
        It will return a list with errors, just in the same
        way of the Protocol.validate function
        """
        errors = []
        
        for name, param in self.iterParams():
            value = protocol.getAttributeValue(name)
            errors += param.validate(value)
        
        return errors
        
    def getParam(self, paramName):
        """Retrieve a param given a the param name
        None is returned if not found
        """        
        return self._paramsDict.get(paramName, None)
    
    def hasParam(self, paramName):
        return paramName in self._paramsDict
        
    def __str__(self):
        s = "Form: \n"
        for section in self.iterSections():
            s += str(section)
        return s
    
    def iterSections(self):
        return self._sectionList
    
    def iterAllParams(self):
        """ Iter all parameters, including ElementGroups. """
        return self._paramsDict.items()
    
    def iterParams(self):
        """ Iter parameters disregarding the ElementGroups. """
        for k, v in self._paramsDict.items():
            if not isinstance(v, ElementGroup):
                yield k, v
        
    def iterPointerParams(self):
        for paramName, param in self._paramsDict.items():
            if isinstance(param, PointerParam):
                yield paramName, param

    def addGeneralSection(self):
        self.addSection(label='General')
        self.addParam('runName', StringParam, label="Run name:", important=True, 
                      help='Select run name label to identify this run.')
        self.addParam('runMode', EnumParam, choices=['resume', 'restart'],
                      label="Run mode", display=EnumParam.DISPLAY_COMBO, default=0,
                      help='The <resume> mode will try to start the execution'
                           'from the last successfully finished step if possible.'
                           'On the contrary, <restart> will delete all previous results'
                           'of this particular run and start from the beginning. This option'
                           'should be used carefully.'
                      )
  
    def addParallelSection(self, threads=1, mpi=8, condition="",
                           hours=72, jobsize=0):

        self.addSection(label='Parallelization')
        self.addParam('hostName', StringParam, default="localhost",
                      label='Execution host',
                      help='Select in which of the available do you want to launch this protocol.')
        if threads > 0:
            self.addParam('numberOfThreads', IntParam, default=threads,
                          label='Threads',
                          help='This option provides shared-memory parallelization on multi-core machines.'
                                'It does not require any additional software, other than <Xmipp>')
        if mpi > 0:
            self.addParam('numberOfMpi', IntParam, default=mpi,
                          label='MPI processes',
                          help='This option provides the number of independent processes spawned'
                                'in parallel by <mpirun> command in a cluster, usually through'
                                'a queue system. This will require that you have compile <Xmipp>'
                                'with <mpi> support.')
        if jobsize > 0:
            self.addParam('mpiJobSize', IntParam, default=jobsize,
                          label='MPI job size', condition="numberOfMpi>1",
                          help='Minimum size of jobs in mpi processes.'
                               'Set to 1 for large images (e.g. 500x500)'
                               'and to 10 for small images (e.g. 100x100)')


class StringParam(Param):
    """Param with underlying String value"""
    def __init__(self, **args):
        Param.__init__(self, paramClass=String, **args)


class TextParam(StringParam):
    """Long string params"""
    def __init__(self, **args):
        StringParam.__init__(self, **args)
        self.height = args.get('height', 5)
        self.width = args.get('width', 30)        
        

class RegexParam(StringParam):
    """Regex based string param"""
    pass


class PathParam(StringParam):
    """Param for path strings"""
    pass


# TODO: Handle filter pattern
class FileParam(PathParam):
    """Filename path"""
    pass


class FolderParam(PathParam):
    """Folder path"""
    pass


class LabelParam(StringParam):
    """ Just the same as StringParam, but to be rendered
    as a label and can not be directly edited by the user
    in the Protocol Form.
    """
    pass

        
class IntParam(Param):
    def __init__(self, **args):
        Param.__init__(self, paramClass=Integer, **args)
        self.addValidator(Format(int, error="should be an integer",
                                 allowsNull=args.get('allowsNull', False)))
        
        
class EnumParam(IntParam):
    """Select from a list of values, separated by comma"""
    # Possible values for display
    DISPLAY_LIST = 0
    DISPLAY_COMBO = 1
    DISPLAY_HLIST = 2  # horizontal list, save space
    
    def __init__(self, **args):
        IntParam.__init__(self, **args)
        self.choices = args.get('choices', [])
        self.display = Integer(args.get('display', EnumParam.DISPLAY_COMBO))
    
    
class FloatParam(Param):
    def __init__(self, **args):
        Param.__init__(self, paramClass=Float, **args)
        self.addValidator(Format(float, error="should be a float",
                                 allowsNull=args.get('allowsNull', False)))

        
class BooleanParam(Param):
    def __init__(self, **args):
        Param.__init__(self, paramClass=Boolean, **args)
        self.addValidator(NonEmptyBool)


class HiddenBooleanParam(BooleanParam):
    def __init__(self, **args):
        Param.__init__(self, paramClass=Boolean, **args)

        
class PointerParam(Param):
    """ This type of Param will serve to select existing objects
    in the database that will be input for some protocol.
    """
    def __init__(self,  paramClass=Pointer, **args):
        Param.__init__(self, paramClass=paramClass, **args)
        # This will be the class to be pointed
        pointerClass = args.get('pointerClass')
        if ',' in pointerClass:
            self.pointerClass = CsvList()
            self.pointerClass.set(pointerClass)
        else:
            self.pointerClass = String(pointerClass) 

        # Some conditions on the pointed candidates
        self.pointerCondition = String(args.get('pointerCondition', None))
        self.allowsNull = Boolean(args.get('allowsNull', False))
        
    def setPointerClass(self, newPointerClass):
        self.pointerClass.set(newPointerClass)


class MultiPointerParam(PointerParam):
    """ This type of Param will serve to select objects
    with DIFFERENT types from the database to be input for some protocol.
    """
    def __init__(self, **args):
        PointerParam.__init__(self, paramClass=PointerList, **args)
        self.maxNumObjects = Integer(args.get('maxNumObjects', 100))
        self.minNumObjects = Integer(args.get('minNumObjects', 2))

        
class RelationParam(Param):
    """ This type of Param is very similar to PointerParam, since it will
    hold a pointer to another object. But, in the PointerParam, we search
    for objects of some Class (maybe with some conditions).
    Here, we search for objects related to a given attribute of a protocol
    by a given relation.
    """
    def __init__(self, **args):
        Param.__init__(self, paramClass=Pointer, **args)
        # This will be the name of the relation
        self._relationName = String(args.get('relationName'))
        # We will store the attribute name in the protocol to be 
        # used as the object for which relations will be search
        self._attributeName = String(args.get('attributeName'))
        # This specify if we want to search for childs or parents
        # of the given attribute of the protocol
        self._direction = Integer(args.get('direction', RELATION_CHILDS))
        self.allowsNull = Boolean(args.get('allowsNull', False))
        
    def getName(self):
        return self._relationName.get()
    
    def getAttributeName(self):
        return self._attributeName.get()
    
    def getDirection(self):
        return self._direction.get()       
        
        
class ProtocolClassParam(StringParam):
    def __init__(self, **args):
        StringParam.__init__(self, **args)
        self.protocolClassName = String(args.get('protocolClassName'))
        self.allowSubclasses = Boolean(args.get('allowSubclasses', False))
        
        
class DigFreqParam(FloatParam):
    """ Digital frequency param. """
    def __init__(self, **args):
        FloatParam.__init__(self, **args)
        self.addValidator(FreqValidator)
        
        
class NumericListParam(StringParam):
    """ This class will serve to have list representations as strings.
     Possible notation are:
     1000 10 1 1 -> to define a list with 4 values [1000, 10, 1, 1], or
     10x2 5x3    -> to define a list with 5 values [10, 10, 5, 5, 5]
     If you ask for more elements than in the list, the last one is repeated
    """
    def __init__(self, **args):
        StringParam.__init__(self, **args)
        self.addValidator(NumericListValidator())
        
        
class NumericRangeParam(StringParam):
    """ This class will serve to specify range of numbers with a string representation.
     Possible notation are:
        "1,5-8,10" -> [1,5,6,7,8,10]
        "2,6,9-11" -> [2,6,9,10,11]
        "2 5, 6-8" -> [2,5,6,7,8]
    """
    def __init__(self, **args):
        StringParam.__init__(self, **args)
        # TODO: ADD a syntax validator
        
        
class TupleParam(Param):
    """ This class will condense a tuple of several
    other params of the same type and related concept.
    For example: min and max, low and high.
    """
    def __init__(self, **args):
        Param.__init__(self, **args)


# -----------------------------------------------------------------------------
#         Validators
# -----------------------------------------------------------------------------
class Validator(object):
    pass


class Conditional(Validator):
    """ Simple validation based on a condition. 
    If the value doesn't meet the condition,
    the error will be returned.
    """
    def __init__(self, error, allowsNull=False):
        self.error = error
        self._allowsNull = allowsNull
        
    def __call__(self, value):
        errors = []
        if value is not None or not self._allowsNull:
            if not self._condition(value):
                errors.append(self.error)
        return errors   
    
    
class Format(Conditional):
    """ Check if the format is right. """
    def __init__(self, valueType, error='Value have not a correct format',
                 allowsNull=False):
        Conditional.__init__(self, error, allowsNull)
        self.valueType = valueType
        
    def _condition(self, value):
        try:
            self.valueType(value)
            return True
        except Exception:
            return False


class NonEmptyCondition(Conditional):
    def __init__(self, error='Value cannot be empty'):
        Conditional.__init__(self, error)
        self._condition = lambda value: len(value) > 0
        
        
class LT(Conditional):
    def __init__(self, threshold,
                 error='Value should be less than the threshold'):
        Conditional.__init__(self, error)
        self._condition = lambda value: value < threshold
        
        
class LE(Conditional):
    def __init__(self, threshold,
                 error='Value should be less or equal than the threshold'):
        Conditional.__init__(self, error)
        self._condition = lambda value: value <= threshold
        
        
class GT(Conditional):
    def __init__(self, threshold,
                 error='Value should be greater than the threshold'):
        Conditional.__init__(self, error)
        self._condition = lambda value: value > threshold


class GE(Conditional):
    def __init__(self, thresold, error='Value should be greater or equal than the threshold'):
        Conditional.__init__(self, error)
        self._condition = lambda value: value >= thresold               


class Range(Conditional):
    def __init__(self, minValue, maxValue, error='Value is outside range'):
        Conditional.__init__(self, error)
        self._condition = lambda value: minValue <= value <= maxValue
        
        
class NumericListValidator(Conditional):
    """ Validator for ListParam. See ListParam. """
    def __init__(self, error='Incorrect format for numeric list param. '):
        Conditional.__init__(self, error)
        
    def _condition(self, value):
        try:
            value = value.replace('x', '')
            parts = value.split()
            for p in parts:
                float(p)
            return True
        except Exception:
            return False    


class NonEmptyBoolCondition(Conditional):
    def __init__(self, error='Boolean param needs to be set.'):
        Conditional.__init__(self, error)
        self._condition = lambda value: value is not None


# --------- Some constants validators ---------------------

Positive = GT(0.0, error='Value should be greater than zero')

FreqValidator = Range(0.001, 0.5,
                      error="Digital frequencies should be between 0.001 and 0.5")

NonEmpty = NonEmptyCondition()
NonEmptyBool = NonEmptyBoolCondition()



# -*- coding: utf-8 -*-
# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *
# * [1] SciLifeLab, Stockholm University
# *
# * This program is free software: you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation, either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program.  If not, see <https://www.gnu.org/licenses/>.
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
"""
Definition of Mock protocols to be used within the tests in the Mock Domain
"""

import pyworkflow.utils as pwutils
import pyworkflow.object as pwobj
import pyworkflow.protocol as pwprot


class SleepingProtocol(pwprot.Protocol):
    def __init__(self, **args):
        pwprot.Protocol.__init__(self, **args)
        self.name = pwobj.String(args.get('name', None))
        self.numberOfSleeps = pwobj.Integer(args.get('n', 1))
        self.runMode = pwobj.Integer(pwprot.MODE_RESUME)

    def sleepStep(self, t, s):
        log = self._getPath("step_%02d.txt" % t)
        import time
        import datetime
        f = open(log, 'w+')
        f.write("Going to sleep at %s\n"
                % pwutils.dateStr(datetime.datetime.now(), True))
        time.sleep(t)
        f.write("  Slept: %d seconds\n" % t)
        f.write("Awakened at %s\n"
                % pwutils.dateStr(datetime.datetime.now(), True))
        f.close()
        return [log]

    def _insertAllSteps(self):
        print("Inserting all steps...")
        for i in range(self.numberOfSleeps.get()):
            self._insertFunctionStep('sleepStep', i + 1, 'sleeping %d' % i)


class ParallelSleepingProtocol(SleepingProtocol):
    def _insertAllSteps(self):
        step1 = self._insertFunctionStep('sleepStep', 1, '1')
        n = 2
        deps = [step1]
        for i in range(n):
            self._insertFunctionStep('sleepStep')


class ProtOutputTest(pwprot.Protocol):
    """ Protocol to test scalar output and input linking"""
    _label = 'test output'

    def __init__(self, **args):
        pwprot.Protocol.__init__(self, **args)
        self.name = pwobj.String(args.get('name', None))

    def _defineParams(self, form):

        section = form.addSection("Input")
        section.addParam('iBoxSize', pwprot.IntParam, allowsPointers=True,
                         default=10,
                         label='Input box size as Integer',
                         validators=[pwprot.Positive])

        section.addParam('nullableInteger', pwprot.IntParam, allowsPointers=True,
                         label='Nullable Integer', allowsNull=True)

    def _createOutputStep(self):
        # New Output would be an Integer
        boxSize = pwobj.Integer(10)

        if self.iBoxSize.hasValue():
            boxSize.set(2*int(self.iBoxSize.get()))

        self._defineOutputs(oBoxSize=boxSize)

    def _insertAllSteps(self):
        self._insertFunctionStep('_createOutputStep')


class ProtMultiPointerTest(pwprot.Protocol):
    """ Class to test how multipointer params are exported to json"""
    def _defineParams(self, form):

        # This should cover Multipointer params that points to attributes...
        # therefore extended attribute of pointers should be used
        form.addParam('mpToAttr', pwprot.MultiPointerParam,
                      label="Multipointer to attribute",
                      pointerClass='String',
                      help="Should point to String inside another protocol")

        # This should cover Multipointer params that points to protocols...
        # therefore extended attribute of pointers should NOT be used
        form.addParam('mpToProts', pwprot.MultiPointerParam,
                      label="Multipointer to sets",
                      pointerClass='Protocol',
                      help="Should point to another protocol")



# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
"""
This module trace through printing 
module and class functions...
Code from: http://wordaligned.org/articles/echo
"""


import inspect
import sys
    

def name(item):
    """ Return an item's name. """
    return item.__name__


def is_classmethod(instancemethod):
    """ Determine if an instancemethod is a classmethod. """
    return instancemethod.im_self is not None


def is_class_private_name(name):
    """ Determine if a name is a class private name. """
    # Exclude system defined names such as __init__, __add__ etc
    return name.startswith("__") and not name.endswith("__")


def method_name(method):
    """ Return a method's name.
    
    This function returns the name the method is accessed by from
    outside the class (i.e. it prefixes "private" methods appropriately).
    """
    mname = name(method)
    if is_class_private_name(mname):
        mname = "_%s%s" % (name(method.im_class), mname)
    return mname


def format_args(args):
    print(args)


def echo(fn, write=sys.stdout.write):
    print(fn)


def echo_instancemethod(klass, method, write=sys.stdout.write):
    """ Change an instancemethod so that calls to it are echoed.
    
    Replacing a classmethod is a little more tricky.
    See: http://www.python.org/doc/current/ref/types.html
    """
    mname = method_name(method)
    never_echo = "__str__", "__repr__",  # Avoid recursion printing method calls
    if mname in never_echo:
        pass
    elif is_classmethod(method):
        setattr(klass, mname, classmethod(echo(method.im_func, write)))
    else:
        setattr(klass, mname, echo(method, write))


def echo_class(klass, write=sys.stdout.write):
    """ Echo calls to class methods and static functions
    """
    for _, method in inspect.getmembers(klass, inspect.ismethod):
        echo_instancemethod(klass, method, write)
    for _, fn in inspect.getmembers(klass, inspect.isfunction):
        setattr(klass, name(fn), staticmethod(echo(fn, write)))


def echo_module(mod, write=sys.stdout.write):
    """ Echo calls to functions and methods in a module.
    """
    for fname, fn in inspect.getmembers(mod, inspect.isfunction):
        setattr(mod, fname, echo(fn, write))
    for _, klass in inspect.getmembers(mod, inspect.isclass):
        echo_class(klass, write)
